from dataclasses import dataclass

import click

from contxt.auth.cli import CliAuth
from contxt.services import (
    ContxtAccessService,
    ContxtDeploymentService,
    ContxtService,
    EmsService,
    EventsService,
    IotService,
    NionicService,
)
from contxt.utils import cachedproperty


@dataclass
class Clients:
    """Holds a user and all client API's"""

    env: str
    org_slug: str

    @cachedproperty
    def org_id(self) -> str:
        if not self.org_slug:
            # Check if user only has access to a single org
            token = self.contxt.session.auth.token_provider.decoded_access_token
            orgs = token.get("organizations", [])
            if len(orgs) != 1:
                raise click.ClickException("Organization required. Try again with `contxt --org=...`.")
            return orgs[0]
        for org in self.contxt.get("organizations"):
            if org["slug"] == self.org_slug:
                return org["id"]
        raise click.ClickException(f"Organization {self.org_slug!r} does not exist.")

    @cachedproperty
    def auth(self) -> CliAuth:
        return CliAuth(env=self.env)

    @cachedproperty
    def contxt(self) -> ContxtService:
        return ContxtService(auth=self.auth, env=self.env)

    @cachedproperty
    def contxt_deployments(self) -> ContxtDeploymentService:
        return ContxtDeploymentService(auth=self.auth, env=self.env)

    @cachedproperty
    def contxt_access(self) -> ContxtAccessService:
        return ContxtAccessService(auth=self.auth, env=self.env)

    @cachedproperty
    def ems(self) -> EmsService:
        return EmsService(auth=self.auth, env=self.env, org_id=self.org_id, org_slug=self.org_slug)

    @cachedproperty
    def events(self) -> EventsService:
        return EventsService(auth=self.auth, env=self.env, org_slug=self.org_slug)

    @cachedproperty
    def nionic(self) -> NionicService:
        return NionicService(auth=self.auth, env=self.env, org_slug=self.org_slug)

    @cachedproperty
    def iot(self) -> IotService:
        return IotService(auth=self.auth, env=self.env, org_slug=self.org_slug)



from csv import DictReader
from typing import IO

import click

from contxt.cli.clients import Clients
from contxt.cli.utils import Serializer


@click.group()
def facilities() -> None:
    """Facilities."""


@facilities.command()
@click.pass_obj
def get(clients: Clients) -> None:
    """Get facilities"""
    facilities = clients.nionic.get_facilities()
    print(Serializer.to_table(facilities))


@facilities.command()
@click.option("--input", required=True, type=click.File(), help="CSV of facilities to create")
@click.pass_obj
def create(clients: Clients, input: IO[str]) -> None:
    """Create facilities from a CSV file, containing the columns:

    \b
    - name: string
    - timezone_name: IANA timezone database name (ex: America/New_York)
    """
    # Parse file
    try:
        facilities = [
            dict(name=r["name"], slug=r["slug"], timezone_name=r["timezone_name"])
            for r in DictReader(input)
        ]
    except KeyError:
        raise click.ClickException("The following columns are required: name, slug, timezone_name")

    # Create facilities
    with click.progressbar(
        facilities,
        label="Creating facilities",
        item_show_func=lambda f: f"Facility {f['name']}" if f else "",
    ) as facilities_:
        for f in facilities_:
            clients.nionic.create_facility(data=f)



"""Main CLI"""

import logging
from pathlib import Path

import click
from requests import RequestException

from contxt import __version__
from contxt.cli.clients import Clients
from contxt.cli.log import init as init_logger

logger = logging.getLogger()
COMMAND_DIR = Path(__file__).parent / "commands"


class Cli(click.MultiCommand):
    def list_commands(self, ctx):
        return sorted(p.stem.replace("_", "-") for p in COMMAND_DIR.glob("*.py"))

    def get_command(self, ctx, name):
        name = name.replace("-", "_")
        f = COMMAND_DIR / f"{name}.py"
        if not f.exists():
            return None
        ns = {}
        code = compile(f.read_text(), f, "exec")
        eval(code, ns, ns)
        return ns[name]

    def __call__(self, *args, **kwargs):
        try:
            return self.main(*args, **kwargs)
        except RequestException as e:
            logger.error(f"{e} (response: {e.response.content})")
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")
            raise


@click.group(cls=Cli, name="contxt", context_settings=dict(help_option_names=["-h", "--help"]))
@click.option(
    "--env",
    type=click.Choice(["production", "staging"]),
    default="production",
    envvar="CONTXT_ENV",
    help="Environment for all API's",
)
@click.option("--org", envvar="CONTXT_ORG", help="Organization slug")
@click.option("-v", "--verbose", count=True, help="Increase verbosity")
@click.version_option(__version__, "-V", "--version")
@click.pass_context
def cli(ctx: click.Context, env: str, org: str, verbose: int) -> None:
    """Contxt CLI"""
    init_logger(verbose)
    ctx.obj = Clients(env=env, org_slug=org)


if __name__ == "__main__":
    cli()



from collections import defaultdict
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List

import click

from contxt.cli.clients import Clients
from contxt.cli.utils import (
    LAST_WEEK,
    NOW,
    ClickPath,
    csv_callback,
    fields_option,
    print_table,
    sort_option,
)
from contxt.models.ems import MainService, ResourceType
from contxt.models.iot import Window
from contxt.utils.serializer import Serializer


@click.group()
def ems() -> None:
    """Energy Management System."""


@ems.command()
@click.argument("facility_id")
@click.option("--resource-type", type=ResourceType, default="electric", help="Resource type")
@fields_option(default="id, name, resource_type", obj=MainService)
@sort_option(default="id")
@click.pass_obj
def mains(
    clients: Clients, facility_id: int, resource_type: ResourceType, fields: List[str], sort: str
) -> None:
    """Get main services"""
    results = clients.nionic.get_main_services(facility_id=facility_id, resource_type=resource_type)
    items = [
        MainService(
            x.id, x.facility_id, x.name, x.type, x.demand.id, x.usage.id, x.created_at, x.updated_at
        )
        for x in results
    ]
    print_table(items=items, keys=fields, sort_by=sort)


@ems.command()
@click.argument("facility_id")
@click.option("--resource-type", type=ResourceType, default="electric", help="Resource type")
@click.option("--start", type=click.DateTime(), help="Start time", required=True)
@click.option("--end", type=click.DateTime(), help="End time", required=True)
@click.pass_obj
def main_data(
    clients: Clients, facility_id: int, resource_type: ResourceType, start: datetime, end: datetime
) -> None:
    """Get main service data"""
    data: Dict[datetime, Dict[str, Any]] = defaultdict(dict)
    services = clients.nionic.get_main_services(facility_id=facility_id, resource_type=resource_type)
    with click.progressbar(
        services,
        label="Downloading main service data",
        item_show_func=lambda s: f"Service {s.name}" if s else "",
    ) as services_:
        for service in services_:
            for t, v in clients.nionic.get_data_point_data(
                data_source_name=service.usage.data_source_name,
                name=service.usage.name,
                start=start.astimezone().isoformat(),
                end=end.astimezone().isoformat(),
                window=Window.MINUTELY,
                per_page=5000,
            ):
                if t in data and service.usage.name in data[t]:
                    data[t][service.usage.name] = v + data[t][service.usage.name]
                else:
                    data[t][service.usage.name] = v

    # Dump
    print_table(items=data)


@ems.command()
@click.argument("facility_id")
@click.option("--resource-type", type=ResourceType, default="electric", help="Resource type")
@click.option("--start", type=click.DateTime(), default=LAST_WEEK.isoformat(), help="Start time")
@click.option("--end", type=click.DateTime(), default=NOW.isoformat(), help="End time")
@click.option(
    "--interval", type=click.Choice(["daily", "monthly"]), default="monthly", help="Time interval"
)
@click.pass_obj
def spend(
    clients: Clients,
    facility_id: int,
    resource_type: ResourceType,
    start: datetime,
    end: datetime,
    interval: str,
) -> None:
    """Get utility spend"""
    items = clients.ems.get_monthly_utility_spend(
        facility_id=facility_id, resource_type=resource_type, start_date=start, end_date=end
    )
    print_table(items=items)


@ems.command()
@click.argument("facility_id")
@click.option("--resource-type", type=ResourceType, default="electric", help="Resource type")
@click.option("--start", type=click.DateTime(), default=LAST_WEEK.isoformat(), help="Start time")
@click.option("--end", type=click.DateTime(), default=NOW.isoformat(), help="End time")
@click.option(
    "--interval", type=click.Choice(["daily", "monthly"]), default="daily", help="Time interval"
)
@click.pass_obj
def usage(
    clients: Clients,
    facility_id: int,
    resource_type: ResourceType,
    start: datetime,
    end: datetime,
    interval: str,
) -> None:
    """Get utility usage"""
    usage = clients.ems.get_usage(
        facility_id=facility_id, resource_type=resource_type, start=start, end=end, interval=interval
    )
    print_table(items=usage.values)


@ems.command()
@click.argument("facility_ids", nargs=-1, type=click.INT)
@click.option(
    "--include",
    default="all",
    callback=csv_callback(options=["mains", "usage"]),
    help="Data to export",
)
@click.option("--start", type=click.DateTime(), default=LAST_WEEK.isoformat(), help="Start time")
@click.option("--end", type=click.DateTime(), default=NOW.isoformat(), help="End time")
@click.option("--output", type=ClickPath(file_okay=False), default=".", help="Path for output")
@click.pass_obj
def export(
    clients: Clients,
    facility_ids: List[int],
    include: List[str],
    start: datetime,
    end: datetime,
    output: Path,
) -> None:
    """Export data for facilities"""
    with click.progressbar(
        facility_ids, label="Downloading data", item_show_func=lambda f: f"Facility {f}" if f else ""
    ):
        for facility in clients.nionic.get_facilities():
            if facility.id not in facility_ids:
                continue
            fpath = output / facility.slug

            # Utility usage
            if "usage" in include:
                usage = clients.ems.get_usage(
                    facility_id=facility.id, start=start, end=end, interval="daily"
                )
                Serializer.to_csv(usage.values, fpath / "ems" / "usage.csv")

            # Main service data
            if "mains" in include:
                data: Dict[datetime, Dict[str, Any]] = defaultdict(dict)
                services = clients.nionic.get_main_services(facility_id=facility.id)
                for service in services:
                    for t, v in clients.nionic.get_data_point_data(
                        data_source_name=service.usage.data_source_name,
                        name=service.usage.name,
                        start=start.astimezone().isoformat(),
                        end=end.astimezone().isoformat(),
                        window=Window.MINUTELY,
                        per_page=5000,
                    ):
                        if t in data and service.usage.name in data[t]:
                            data[t][service.usage.name] = v + data[t][service.usage.name]
                        else:
                            data[t][service.usage.name] = v
                Serializer.to_csv(data, fpath / "ems" / "main_service_usage.csv")

    print(f"Wrote data to {output}")



from dataclasses import dataclass
from datetime import date, datetime
from enum import Enum
from typing import ClassVar, List, Optional

from . import ApiField, ApiObject, Formatters, Parsers
from .events import Event


class ResourceType(Enum):
    COMBINED = "combined"
    ELECTRIC = "electric"
    GAS = "gas"
    WATER = "water"


@dataclass
class MainService(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id", data_type=int),
        ApiField("facility_id", data_type=int),
        ApiField("name"),
        ApiField("type", attr_key="resource_type", data_type=ResourceType),
        ApiField("demand_datapoint_id", data_type=int),
        ApiField("usage_datapointid", data_type=int),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    id: int
    facility_id: int
    name: str
    resource_type: ResourceType
    demand_datapoint_id: int
    usage_datapoint_id: int
    created_at: datetime
    updated_at: datetime


@dataclass
class Metric(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("label", creatable=True, updatable=True),
        ApiField("description", creatable=True, updatable=True),
        ApiField("organization_id", creatable=True),
        ApiField("time_interval", creatable=True, updatable=True),
        ApiField("units", creatable=True, updatable=True),
        ApiField("is_global", data_type=bool, creatable=True),
        ApiField("is_calculated", data_type=bool),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    label: str
    description: str
    organization_id: str
    time_interval: str
    units: str
    id: Optional[str] = None
    is_global: Optional[bool] = None
    is_calculated: Optional[bool] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None

    @property
    def normalized_label(self) -> str:
        return Formatters.normalize_label(self.label)


@dataclass
class MetricValue(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id", optional=True),
        ApiField("effective_start_date", data_type=Parsers.datetime, creatable=True, updatable=True),
        ApiField("effective_end_date", data_type=Parsers.datetime, creatable=True, updatable=True),
        ApiField("value", data_type=float, creatable=True, updatable=True),
        ApiField("created_at", data_type=Parsers.datetime, optional=True),
        ApiField("updated_at", data_type=Parsers.datetime, optional=True),
        ApiField("is_estimated", data_type=bool, optional=True),
    )

    effective_start_date: datetime
    effective_end_date: datetime
    value: str
    id: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    is_estimated: Optional[bool] = None


@dataclass
class UtilityPeriod(ApiObject):
    _api_fields: ClassVar = (
        ApiField("event_time", data_type=Parsers.datetime),
        ApiField("value", data_type=int),
    )

    event_time: datetime
    value: int


UtilitySpendPeriod = UtilityPeriod
UtilityUsagePeriod = UtilityPeriod


@dataclass
class UtilitySpend(ApiObject):
    _api_fields: ClassVar = (
        ApiField("type"),
        ApiField("currency"),
        ApiField("values", data_type=UtilitySpendPeriod),
    )

    type: str
    currency: str
    values: List[UtilitySpendPeriod]

    @property
    def periods(self):
        return self.values


@dataclass
class UtilityUsage(ApiObject):
    _api_fields: ClassVar = (
        ApiField("type"),
        ApiField("unit"),
        ApiField("values", data_type=UtilityUsagePeriod),
    )

    type: str
    unit: str
    values: List[UtilityUsagePeriod]

    @property
    def periods(self):
        return self.values


@dataclass
class UtilityContractReminder(ApiObject):
    _api_fields: ClassVar = (
        ApiField("utility_contract_id", data_type=int),
        ApiField("user_id", data_type=str),
        ApiField("user_event_subscription_id", data_type=str),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    utility_contract_id: int
    user_id: str
    user_event_subscription_id: str
    created_at: datetime
    updated_at: datetime


@dataclass
class UtilityContract(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id", data_type=int),
        ApiField("name"),
        ApiField("facility_id", data_type=int),
        ApiField("status"),
        ApiField("start_date", data_type=Parsers.date),
        ApiField("end_date", data_type=Parsers.date),
        ApiField("rate_narrative"),
        ApiField("file_id"),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
        ApiField("created_by"),
        ApiField("utility_contract_reminders", data_type=UtilityContractReminder),
        ApiField("report_event_id"),
        ApiField("report_event", data_type=Event),
    )

    id: int
    name: str
    facility_id: int
    status: str
    start_date: datetime
    end_date: datetime
    rate_narrative: str
    file_id: str
    created_at: datetime
    updated_at: datetime
    created_by: str
    utility_contract_reminders: List[UtilityContractReminder]
    report_event_id: str
    report_event: Event


@dataclass
class UtilityStatement(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id", data_type=int),
        ApiField("utility_meter_id", data_type=int),
        ApiField("statement_month", data_type=str),
        ApiField("statement_year", data_type=str),
        ApiField("interval_start", data_type=Parsers.date),
        ApiField("interval_end", data_type=Parsers.date),
        ApiField("file_id", data_type=int),
        ApiField("is_validated", data_type=bool),
        ApiField("pm_id", data_type=int),
        ApiField("uploaded_to_pm", data_type=bool),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    id: int
    utility_meter_id: int
    statement_month: str
    statement_year: str
    interval_start: date
    interval_end: date
    file_id: int
    is_validated: bool
    pm_id: int
    uploaded_to_pm: bool
    created_at: datetime
    updated_at: datetime


@dataclass
class UtilityMeter(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id", data_type=int),
        ApiField("label", data_type=str),
        ApiField("utility_account_id", data_type=int),
        ApiField("building_id", data_type=int),
        ApiField("service_type", data_type=str),
        ApiField("logical_meter_id", data_type=str),
        ApiField("meter_number", data_type=str),
        ApiField("active_start", data_type=Parsers.date),
        ApiField("active_end", data_type=Parsers.date),
        ApiField("pm_id", data_type=int),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    id: int
    label: str
    utility_account_id: int
    building_id: int
    service_type: str
    logical_meter_id: str
    meter_number: str
    active_start: date
    active_end: date
    pm_id: int
    created_at: datetime
    updated_at: datetime


@dataclass
class UtilityAccount(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id", data_type=int),
        ApiField("logical_account_id", data_type=str),
        ApiField("facility_id", data_type=int),
        ApiField("account_number", data_type=str),
        ApiField("label", data_type=str),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    id: int
    logical_account_id: int
    facility_id: int
    account_number: str
    label: str
    created_at: datetime
    updated_at: datetime



from typing import List, Optional

import click
import yaml
from requests.exceptions import HTTPError

from contxt.cli.clients import Clients
from contxt.cli.utils import fields_option, print_item, print_table, sort_option
from contxt.models.contxt import Cluster


@click.group()
def clusters() -> None:
    """Clusters"""


@clusters.command()
@click.argument("cluster_slug", required=False)
@fields_option(default="id, host, slug", obj=Cluster)
@sort_option(default="id")
@click.pass_obj
def get(clients: Clients, fields: List[str], sort: str, cluster_slug: Optional[str]) -> None:
    """Get clusters"""
    (items, fields) = (
        ([clients.contxt_deployments.get_cluster(clients.org_id, cluster_slug)], None)  # type: ignore
        if cluster_slug
        else (clients.contxt_deployments.get_clusters(clients.org_id), fields)
    )
    print_table(items=items, keys=fields, sort_by=sort)


@clusters.command()
@click.argument("slug")
@click.pass_obj
def login(clients: Clients, slug: str) -> None:
    """Get cluster config"""
    try:
        clusters = clients.contxt_deployments.get_clusters(clients.org_id)
        cluster_host = next((cluster.host for cluster in clusters if cluster.slug == slug), None)
        if cluster_host is None:
            raise click.ClickException(f"Unable to locate cluster with the slug {slug}")

        config = clients.auth.get_cluster_config(cluster_host)
        print(yaml.safe_dump(config))
    except HTTPError:
        pass


@clusters.command()
@click.option("--host", prompt=True)
@click.option(
    "--slug",
    prompt=True,
    help="The slugified name you would like to use for this cluster."
    " You will need to reference this in other commands so it's "
    "ideal to make this value easy to remember",
)
@click.option(
    "--description",
    prompt=True,
    help="Information about what this cluster is for and what environment it belongs to",
)
@click.option(
    "--environment-type",
    type=click.Choice(["production", "nonproduction", "blended"]),
    default="production",
    prompt=True,
)
@click.option(
    "--certificate-authority",
    type=str,
    prompt=True,
    help="Base64 encoded certificate authority (CA) for the cluster",
)
@click.pass_obj
def register(
    clients: Clients,
    description: str,
    slug: str,
    host: str,
    environment_type: str,
    certificate_authority: str,
) -> None:
    result = clients.contxt_deployments.post(
        f"{clients.org_id}/clusters",
        json={
            "host": host,
            "slug": slug,
            "description": description,
            "type": "kubernetes",
            "environment_type": environment_type,
            "certificate_authority": certificate_authority,
        },
    )
    print_item(result)


@clusters.command()
@click.argument("curr_slug", metavar="SLUG")
@click.option("--host")
@click.option(
    "--description",
    help="Information about what this cluster is for and what environment it belongs to",
)
@click.option("--environment-type", type=click.Choice(["production", "nonproduction", "blended"]))
@click.option("--certificate-authority", type=str)
@click.pass_obj
def update(
    clients: Clients,
    curr_slug: str,
    host: Optional[str],
    description: Optional[str],
    environment_type: Optional[str],
    certificate_authority: Optional[str],
) -> None:
    """Update a cluster"""
    params = {
        k: v
        for k, v in {
            "host": host,
            "description": description,
            "environment_type": environment_type,
            "certificate_authority": certificate_authority,
        }.items()
        if v is not None
    }
    result = clients.contxt_deployments.put(
        f"{clients.org_id}/clusters/{curr_slug}",
        json=params,
    )
    print_item(result)



from datetime import datetime

import click

from contxt.cli.clients import Clients
from contxt.cli.utils import Serializer


@click.group()
def metrics() -> None:
    """Metrics."""


@metrics.command()
@click.pass_obj
def labels(clients: Clients) -> None:
    """Get Metric Labels"""
    labels = clients.nionic.get_metric_labels()
    print(Serializer.to_table(labels))


@metrics.command()
@click.option(
    "--facility-id",
    required=True,
    help="Facility ID",
)
@click.option("--label", required=True, help="Facility ID")
@click.option("--start", required=True, type=click.DateTime())
@click.pass_obj
def data(clients: Clients, facility_id: int, label: str, start: datetime) -> None:
    """Get metric data"""
    results = clients.nionic.get_metric_data(label, facility_id, start.astimezone().isoformat())
    print(Serializer.to_table(results))



import logging
from collections import defaultdict
from csv import DictReader, DictWriter
from datetime import datetime
from pathlib import Path
from typing import IO, Any, Dict, List, Optional, cast

import click
from requests import HTTPError
from slugify import slugify

from contxt.cli.clients import Clients
from contxt.cli.utils import (
    LAST_WEEK,
    NOW,
    ClickPath,
    fields_option,
    print_table,
    sort_option,
    str_to_bool,
)
from contxt.models.iot import Feed, Field, FieldCategory, FieldGrouping, FieldValueType, Window
from contxt.utils.serializer import Serializer

NEW_FIELD_ATTRS = [
    "field_descriptor",
    "label",
    "value_type",
    "units",
    "is_totalizer",
    "grouping",
    "category",
]


@click.group()
def iot() -> None:
    """Internet-of-Things."""


@iot.group()
def fields() -> None:
    """IoT Fields"""


@iot.group()
def groupings() -> None:
    """IoT Field Groupings"""


@iot.group()
def categories() -> None:
    """IoT Grouping Categories"""


@iot.group()
def data() -> None:
    """IoT Field Data"""


@iot.command()
@fields_option(default="id, key, facility_id", obj=Feed)
@sort_option(default="id")
@click.option(
    "--facility_id",
    required=False,
    type=int,
    help="Filter on Facility ID",
)
@click.pass_obj
def feeds(clients: Clients, fields: List[str], sort: str, facility_id: Optional[int]) -> None:
    """Get feeds"""
    items = clients.iot.get_feeds(facility_id=facility_id)
    print_table(items=items, keys=fields, sort_by=sort)


@iot.command()
@click.argument("feed_key")
@click.argument("enabled", type=bool)
@click.pass_obj
def provisioning_mode(clients: Clients, feed_key: str, enabled: bool) -> None:
    """Enable/Disable provisioning mode for a feed"""
    feed = clients.iot.get_feed_with_key(key=feed_key)
    if not feed:
        raise click.ClickException(f"Feed with key {feed_key} does not exist.")

    clients.iot.update_feed_provisioning_mode(feed.id, enabled)
    newState = "enabled" if enabled else "disabled"
    print(f"Provisioning mode is now {newState}")


@fields.command()
@click.argument("feed_key", type=str)
@fields_option(default="id, feed_key, name, field_human_name", obj=Field)
@sort_option(default="id")
@click.pass_obj
def get(clients: Clients, feed_key: str, fields: List[str], sort: str) -> None:
    """Get fields"""
    feed = clients.iot.get_feed_with_key(key=feed_key)
    if not feed:
        raise click.ClickException(f"Feed with key {feed_key} does not exist.")

    items = clients.iot.get_fields_for_feed(feed.id)
    print_table(items=items, keys=fields, sort_by=sort)


@groupings.command("get")
@click.argument("facility_id", type=int)
@fields_option(default="id, label, slug, description", obj=FieldGrouping)
@sort_option(default="id")
@click.pass_obj
def groupings_get(clients: Clients, facility_id: int, fields: List[str], sort: str) -> None:
    """Get field groupings"""
    items = clients.iot.get_field_groupings_for_facility(facility_id)
    print_table(items=items, keys=fields, sort_by=sort)


@categories.command("get")
@click.argument("facility_id", type=int)
@fields_option(default="id, name, description, parent_category_id", obj=FieldCategory)
@sort_option(default="name")
@click.pass_obj
def categories_get(clients: Clients, facility_id: int, fields: List[str], sort: str) -> None:
    """Get categories for facility"""
    items = clients.iot.get_categories_for_facility(facility_id)
    print_table(items=items, keys=fields, sort_by=sort)


@data.command()
@click.option("--output-id", required=True, type=int, help="Output ID to delete from")
@click.option(
    "--field-human-name", required=True, help="The field to delete from. Ex: power.demand, power.usage"
)
@click.option(
    "--interval",
    required=True,
    type=click.Choice([str(v.value) for v in Window]),
    callback=lambda ctx, param, value: Window(int(value)) if value is not None else None,
    help="Time interval",
)
@click.option(
    "--time",
    required=True,
    type=click.DateTime(),
    help="The timestamp of the datapoint. Ex: 2020-07-21T05:31:00Z",
)
@click.pass_obj
def delete(clients: Clients, output_id: int, field: str, interval: Window, time: datetime) -> None:
    """Delete a data point"""
    clients.iot.delete_time_series_point(output_id, field, interval, time)
    print("Data point deleted")


@data.command("get")
@click.argument("feed_id", type=int)
@click.option("--start", type=click.DateTime(), default=LAST_WEEK.isoformat(), help="Start time")
@click.option("--end", type=click.DateTime(), default=NOW.isoformat(), help="End time")
@click.option(
    "--interval",
    type=click.Choice([str(v.value) for v in Window]),
    default="3600",
    callback=lambda ctx, param, value: Window(int(value)) if value is not None else None,
    help="Time interval",
)
@click.option(
    "--output", type=ClickPath(dir_okay=False, writable=True), default="data.csv", help="Path for output"
)
@click.pass_obj
def data_get(
    clients: Clients, feed_id: str, start: datetime, end: datetime, interval: Window, output: Path
) -> None:
    """Get field data"""
    fields = clients.iot.get_fields_for_feed(feed_id)
    print(f"Fetching iot data for {len(fields)} tags from {start} to {end}")

    data: Dict[datetime, Dict[str, Any]] = defaultdict(dict)
    with click.progressbar(
        fields,
        label="Downloading iot data",
        item_show_func=lambda f: f"Field {f.field_human_name}" if f else "",
    ) as fields_:
        for field in fields_:
            for t, v in clients.iot.get_time_series_for_field(
                field=field, start_time=start, end_time=end, window=interval
            ):
                data[t][field.field_human_name] = v

    # Output to csv
    print(f"Writing data to {output}...")
    flat_data = [{"timestamp": k, **v} for k, v in data.items()]
    Serializer.to_csv(flat_data, output)


@fields.command()
@click.argument("feed_key")
@click.option("--input", required=True, type=click.File(), help="CSV of fields to create")
@click.pass_obj
def create(clients: Clients, feed_key: str, input: IO[str]) -> None:
    """Create (provision) fields"""
    # Get feed
    feed = clients.iot.get_feed_with_key(feed_key)
    if not feed:
        raise click.ClickException(f"Feed with key {feed_key} does not exist.")

    # Parse fields
    try:
        fields = [
            [
                Field(
                    feed_key=feed_key,
                    field_descriptor=r["field_descriptor"],
                    label=r["label"],
                    units=r["units"],
                    value_type=FieldValueType(r["value_type"].lower()),
                    is_hidden=False,
                    is_totalizer=str_to_bool(r["is_totalizer"]),
                ),
                r["grouping"],
                r["category"],
            ]
            for r in DictReader(input)
        ]
    except KeyError:
        raise click.ClickException(f"The following columns are required: {NEW_FIELD_ATTRS}")

    # Provision fields
    curr_fields = {f.field_descriptor: f for f in clients.iot.get_fields_for_feed(feed.id)}
    failures = []
    with click.progressbar([t[0] for t in fields], label="Provisioning fields") as _fields:
        for i, field in enumerate(_fields):
            field = cast(Field, field)
            if field.field_descriptor not in curr_fields:
                # New field, create it
                try:
                    fields[i][0] = clients.iot.provision_field_for_feed(feed.id, field)
                except HTTPError as e:
                    logging.debug(e)
                    failures.append(
                        {"field_descriptor": field.field_descriptor, "error_message": e.response.text}
                    )
            else:
                # Existing field, ignore it
                fields[i][0] = curr_fields[field.field_descriptor]

    print_provisioning_failures(
        message="Some fields could not be provisioned - resolve errors and rerun.",
        failures=failures,
    )

    # Add fields to grouping
    groupings = {g.slug: g for g in clients.iot.get_field_groupings_for_facility(feed.facility_id)}
    failures.clear()
    with click.progressbar(fields, label="Adding fields to groupings") as fields_:
        for field, grouping_label, category in fields_:
            grouping_slug = slugify(cast(str, grouping_label))
            field = cast(Field, field)
            if grouping_slug not in groupings:
                # New grouping, create it
                grouping = clients.iot.create_grouping(
                    facility_id=feed.facility_id,
                    label=grouping_label,
                    description=grouping_label,
                    is_public=True,
                    field_category_id=[],
                )
                groupings[grouping.slug] = grouping
            else:
                # Existing grouping, grab it
                grouping = groupings[grouping_slug]
            try:
                clients.iot.add_field_to_grouping(grouping.id, field.id)
            except HTTPError as e:
                logging.debug(e)
                if e.response.text != '{"code":400,"message":"Field is already part of this grouping"}':
                    failures.append(
                        {
                            "field_id": field.id,
                            "field_descriptor": field.field_descriptor,
                            "grouping_slug": grouping.slug,
                            "error_message": e.response.text,
                        }
                    )

    print_provisioning_failures(
        message="Some fields could not be added to groupings - resolve errors and rerun.",
        failures=failures,
    )

    # Add groupings to categories
    categories = {c.name: c for c in clients.iot.get_categories_for_facility(feed.facility_id)}
    new_groups = {g: c for f, g, c in fields}

    # Verify that all categories exist
    new_categories = list(set(v for v in new_groups.values()))
    missing_categories = [{"category": c} for c in new_categories if c not in categories]
    print_provisioning_failures(
        message=(
            f"Some categories were not found for facility {feed.facility_id} - resolve errors and rerun."
        ),
        failures=missing_categories,
    )

    failures.clear()
    with click.progressbar(new_groups, label="Adding groupings to categories") as _groups:
        for group in _groups:
            grouping_id = groupings[slugify(cast(str, group))].id
            category_id = categories[new_groups[group]].id if new_groups[group] != "" else None
            try:
                clients.iot.add_grouping_to_category(grouping_id, category_id)
            except HTTPError as e:
                logging.debug(e)
                failures.append(
                    {
                        "grouping_id": grouping_id,
                        "category_id": category_id,
                        "error_message": e.response.text,
                    }
                )

    print_provisioning_failures(
        message="Some groupings could not be added to categories - resolve errors and rerun.",
        failures=failures,
    )


def print_provisioning_failures(message: str, failures: List[Dict[str, Any]]) -> None:
    if len(failures) > 0:
        print_table(items=failures)
        raise click.ClickException(message)


@fields.command()
@click.argument("feed_key")
@click.option("--output", type=click.File(mode="w"), help="Path for output")
@click.pass_obj
def unprovisioned(clients: Clients, feed_key: str, output: Optional[IO[str]]) -> None:
    """Get unprovisioned fields"""
    items = clients.iot.get_unprovisioned_fields_for_feed_key(feed_key)
    if not output:
        print_table(items=items or [], keys=["field_descriptor"])
    else:
        writer = DictWriter(output, fieldnames=NEW_FIELD_ATTRS)
        writer.writeheader()
        writer.writerows([{"field_descriptor": f.field_descriptor} for f in items])


@groupings.command("create")
@click.argument("facility_id")
@click.option("--label", required=True)
@click.option("--description", required=True)
@click.option("--is-public", required=True, type=bool)
@click.option("--field-category-id", required=True)
@click.pass_obj
def groupings_create(
    clients: Clients,
    facility_id: int,
    label: str,
    description: str,
    is_public: bool,
    field_category_id: str,
) -> None:
    """Create a new grouping"""
    grouping = clients.iot.create_grouping(
        facility_id=facility_id,
        label=label,
        description=description,
        is_public=is_public,
        field_category_id=[field_category_id],
    )
    print(grouping)


@groupings.command()
@click.argument("grouping_id")
@click.argument("field_id")
@click.pass_obj
def add_field(clients: Clients, grouping_id: str, field_id: str) -> None:
    """Add field to grouping"""
    grouping_field = clients.iot.add_field_to_grouping(grouping_id=grouping_id, field_id=field_id)
    print(grouping_field)


@fields.command("delete")
@click.argument("field_id", nargs=-1)
@click.pass_obj
def fields_delete(clients: Clients, field_id: List[str]):
    """Delete (unprovision) fields"""
    for id in field_id:
        clients.iot.unprovision_field(id)


@groupings.command("delete")
@click.argument("grouping_id", nargs=-1)
@click.pass_obj
def groupings_delete(clients: Clients, grouping_id: List[str]):
    """Delete field groupings (does not delete fields within grouping)"""
    for id in grouping_id:
        clients.iot.delete_field_grouping(id)



from dataclasses import dataclass
from datetime import datetime
from json import loads
from typing import ClassVar, List, Optional

from . import ApiField, ApiObject, Parsers


@dataclass
class ConfigValue(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("key"),
        ApiField("value"),
        ApiField("type"),
        ApiField("configuration_id"),
        ApiField("is_hidden", data_type=bool),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    id: str
    key: str
    value: str
    type: str
    configuration_id: str
    is_hidden: bool
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None

    def __post_init__(self) -> None:
        self.value = self._init_value(self.value, self.type)

    def _init_value(self, value, type):
        if type == "integer":
            return int(value)
        elif type == "json":
            return loads(value)
        else:
            return str(value)


@dataclass
class Config(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("name"),
        ApiField("description"),
        ApiField("environment_id"),
        ApiField("ConfigurationValues", attr_key="config_values", data_type=ConfigValue),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    id: str
    name: str
    description: str
    environment_id: str
    config_values: List[ConfigValue]
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


@dataclass
class OrganizationUser(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("user_id"),
        ApiField("organization_id"),
        ApiField("is_primary", data_type=bool),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    id: str
    user_id: str
    organization_id: str
    is_primary: bool
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


@dataclass
class Organization(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("name", creatable=True),
        ApiField("slug", optional=True),
        ApiField(
            "OrganizationUser", attr_key="organization_user", data_type=OrganizationUser, optional=True
        ),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    name: str
    id: Optional[str] = None
    slug: Optional[str] = None
    organization_user: Optional[OrganizationUser] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


@dataclass
class UserRole(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("user_id"),
        ApiField("role_id"),
        ApiField("mapped_from_external_group", data_type=bool),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    user_id: str
    role_id: str
    mapped_from_external_group: bool
    id: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


@dataclass
class Role(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("name"),
        ApiField("description"),
        ApiField("organization_id"),
        ApiField("UserRole", attr_key="user_role", data_type=UserRole, optional=True),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    name: str
    description: str
    organization_id: str
    user_role: Optional[UserRole] = None
    id: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


@dataclass
class User(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("first_name"),
        ApiField("last_name"),
        ApiField("email"),
        ApiField("is_activated", data_type=bool),
        ApiField("Roles", attr_key="roles", data_type=Role, optional=True),
        ApiField("is_superuser"),
        ApiField("organizations", data_type=Organization, optional=True),
        ApiField("phone_number"),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    first_name: str
    last_name: str
    email: str
    phone_number: str
    is_activated: bool
    is_superuser: bool
    roles: Optional[List[Role]] = None
    organizations: Optional[List[Organization]] = None
    id: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


@dataclass
class ServiceEnvironmentVariable(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("service_id"),
        ApiField("environment"),
        ApiField("key"),
        ApiField("value"),
        ApiField("is_protected"),
        ApiField("config_map_value_id"),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    id: int
    service_id: int
    environment: str
    key: str
    value: str
    is_protected: bool
    config_map_value_id: int
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


@dataclass
class Domain(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("name"),
        ApiField("cert_name"),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    id: int
    name: str
    cert_name: str
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


@dataclass
class Frontend(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("service_id"),
        ApiField("vhost"),
        ApiField("force_ssl"),
        ApiField("domain_id"),
        ApiField("domain", data_type=Domain),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    id: int
    service_id: int
    vhost: str
    force_ssl: bool
    domain_id: int
    domain: Optional[Domain] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


@dataclass
class Image(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("provider"),
        ApiField("image_name"),
        ApiField("image_tag_id"),
        ApiField("image_secret_id"),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    id: int
    provider: str
    image_name: str
    image_tag_id: str
    image_secret_id: int
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


@dataclass
class Service(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id", data_type=int),
        ApiField("name"),
        ApiField("slug"),
        ApiField("type"),
        ApiField("project_id", data_type=int),
        ApiField("description"),
        ApiField("deployment_strategy"),
        ApiField("created_at", data_type=Parsers.datetime),
    )

    id: int
    name: str
    slug: str
    type: str
    project_id: int
    description: str
    deployment_strategy: str
    created_at: Optional[datetime] = None


@dataclass
class ServiceInstance(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id", data_type=int),
        ApiField("name"),
        ApiField("slug"),
        ApiField("description"),
        ApiField("descriptor"),
        ApiField("service_id", data_type=int),
        ApiField("project_environment_id"),
        ApiField("client_id"),
        ApiField("command"),
        ApiField("arguments"),
        ApiField("last_deployed_at"),
        ApiField("last_configured_at"),
        ApiField("service_env_variables", data_type=ServiceEnvironmentVariable, optional=True),
        ApiField("frontend", data_type=Frontend, optional=True),
        ApiField("image", data_type=Image, optional=True),
        ApiField("service_type"),
        ApiField("created_at", data_type=Parsers.datetime),
    )

    id: int
    name: str
    slug: str
    description: str
    descriptor: str
    service_id: int
    project_environment_id: str
    client_id: str
    command: str
    arguments: str
    last_deployed_at: Optional[datetime]
    last_configured_at: Optional[datetime]
    service_type: str
    service_env_variables: Optional[List[ServiceEnvironmentVariable]] = None
    frontend: Optional[Frontend] = None
    image: Optional[Image] = None
    created_at: Optional[datetime] = None


@dataclass
class ServiceInstanceGrant(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("from_service_instance_id"),
        ApiField("to_service_instance_id"),
        ApiField("created_at", data_type=Parsers.datetime),
    )

    id: str
    from_service_instance_id: str
    to_service_instance_id: int
    created_at: Optional[datetime] = None


@dataclass
class Project(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id", data_type=int),
        ApiField("name"),
        ApiField("slug"),
        ApiField("description"),
        ApiField("type"),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("Roles", attr_key="roles", data_type=Role, optional=True),
        ApiField("Services", attr_key="services", data_type=Service, optional=True),
    )

    id: int
    name: str
    slug: str
    description: str
    type: str
    roles: Optional[List[Role]] = None
    services: Optional[List[Service]] = None
    created_at: Optional[datetime] = None


@dataclass
class ProjectEnvironment(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id", data_type=int),
        ApiField("slug"),
        ApiField("project_id", data_type=int),
        ApiField("cluster_id"),
        ApiField("cluster_slug"),
        ApiField("name"),
        ApiField("description"),
        ApiField("type"),
        ApiField("archived"),
        ApiField("created_at", data_type=Parsers.datetime),
    )

    id: int
    slug: str
    project_id: int
    cluster_id: str
    cluster_slug: str
    name: str
    description: str
    type: str
    created_at: Optional[datetime] = None


@dataclass
class EdgeNode(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("name"),
        ApiField("project_id"),
        ApiField("organization_id"),
        ApiField("description"),
        ApiField("client_id"),
        ApiField("created_at", data_type=Parsers.datetime),
    )

    id: str
    name: str
    project_id: int
    organization_id: str
    description: str
    client_id: str
    created_at: Optional[datetime] = None


@dataclass
class EdgeNodeGrant(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("from_edge_node_id"),
        ApiField("to_service_instance_id"),
        ApiField("created_at", data_type=Parsers.datetime),
    )

    id: str
    from_edge_node_id: str
    to_service_instance_id: int
    created_at: Optional[datetime] = None


@dataclass
class Cluster(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("description", creatable=True),
        ApiField("infrastructure_id", creatable=True),
        ApiField("organization_id", creatable=True),
        ApiField("region", creatable=True),
        ApiField("slug", creatable=True),
        ApiField("host", creatable=True),
        ApiField("type", creatable=True),
        ApiField("certificate_authority", creatable=True),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    description: str
    infrastructure_id: int
    organization_id: str
    region: str
    slug: str
    type: str
    host: Optional[str] = None
    id: Optional[str] = None
    certificate_authority: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None



import logging

from click import echo, style

VERBOSITY = {0: "error", 1: "info", 2: "debug"}


class ClickHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        try:
            echo(self.format(record), err=True)
        except Exception:
            self.handleError(record)


class ClickFormattter(logging.Formatter):
    STYLES = {
        "debug": dict(fg="blue"),
        "info": dict(fg="green"),
        "warning": dict(fg="yellow"),
        "error": dict(fg="red"),
        "critical": dict(fg="red"),
        "exception": dict(fg="bright_red"),
    }

    def format(self, record: logging.LogRecord) -> str:
        if not record.exc_info:
            msg = record.getMessage()
            level = record.levelname.lower()
            if level in self.STYLES:
                prefix = style(f"{level.capitalize()}: ", **self.STYLES[level])  # type: ignore
                msg = "\n".join(prefix + x for x in msg.splitlines())
            return msg
        return super().format(record)


def init(level: int) -> None:
    logger = logging.getLogger()
    handler = ClickHandler()
    handler.formatter = ClickFormattter()
    log_level = getattr(logging, VERBOSITY.get(level, "error").upper())
    logger.setLevel(log_level)
    logger.handlers = [handler]
    logger.propagate = False



from typing import List, Optional

import click

from contxt.cli.clients import Clients
from contxt.cli.utils import OPTIONAL_PROMPT_KWARGS, fields_option, print_item, print_table, sort_option
from contxt.models.contxt import ProjectEnvironment


@click.group()
def project_envs() -> None:
    """Projects Environments."""


@project_envs.command()
@click.argument("project_slug")
@click.pass_obj
@fields_option(default="id, slug, name, type, description", obj=ProjectEnvironment)
@sort_option(default="id")
def get(clients: Clients, project_slug: str, fields: List[str], sort: str) -> None:
    """Get project environment(s)"""
    result = clients.contxt_deployments.get(f"{clients.org_id}/projects/{project_slug}/environments")

    if "cluster_slug" in fields:
        clusters = clients.contxt_deployments.get_clusters(clients.org_id)
        for r in result:
            r["cluster_slug"] = next(c.slug for c in clusters if c.id == r["cluster_id"])

    print_table(result, keys=fields, sort_by=sort)


@project_envs.command()
@click.option("--project-slug", prompt=True)
@click.option("--name", prompt=True)
@click.option("--slug", prompt=True)
@click.option("--description", **OPTIONAL_PROMPT_KWARGS)
@click.option(
    "--type", type=click.Choice(["production", "nonproduction"]), default="nonproduction", prompt=True
)
@click.option("--cluster-id", prompt=True)
@click.option(
    "--deployment-strategy",
    type=click.Choice(["contxt-managed", "self-managed"]),
    default="self-managed",
    prompt=True,
)
@click.pass_obj
def create(
    clients: Clients,
    project_slug: str,
    name: str,
    slug: str,
    description: Optional[str],
    type: str,
    cluster_id: str,
    deployment_strategy: str,
) -> None:
    """Create a project environment"""
    result = clients.contxt_deployments.post(
        f"{clients.org_id}/projects/{project_slug}/environments",
        {
            "cluster_id": cluster_id,
            "slug": slug,
            "name": name,
            "type": type,
            "deployment_strategy": deployment_strategy,
            "description": description,
        },
    )
    print_item(result)


@project_envs.command()
@click.argument("project_slug")
@click.argument("curr_slug", metavar="SLUG")
@click.option("--name")
@click.option("--slug")
@click.option("--type")
@click.option("--cluster-id")
@click.option("--deployment-strategy")
@click.option("--description")
@click.pass_obj
def update(
    clients: Clients,
    project_slug: str,
    curr_slug: str,
    cluster_id: Optional[str],
    slug: Optional[str],
    name: Optional[str],
    type: Optional[str],
    deployment_strategy: Optional[str],
    description: Optional[str],
) -> None:
    """Update a project environment"""
    clients.contxt_deployments.put(
        f"{clients.org_id}/projects/{project_slug}/environments/{curr_slug}",
        {
            "name": name,
            "slug": slug,
            "type": type,
            "cluster_id": cluster_id,
            "deployment_strategy": deployment_strategy,
            "description": description,
        },
    )


@project_envs.command()
@click.argument("project_slug")
@click.argument("slug")
@click.pass_obj
def delete(clients: Clients, project_slug: str, slug: str) -> None:
    """Delete a project environment"""
    clients.contxt_deployments.delete(f"{clients.org_id}/projects/{project_slug}/environments/{slug}")



from typing import List

import click

from contxt.cli.clients import Clients
from contxt.cli.utils import fields_option, print_item, print_table, sort_option
from contxt.models.contxt import Role


@click.group("roles")
def roles() -> None:
    """Roles."""


@roles.command()
@fields_option(default="id, name, description", obj=Role)
@sort_option(default="id")
@click.pass_obj
def get(clients: Clients, fields: List[str], sort: str) -> None:
    """Get roles for a project"""
    items = clients.contxt_access.get_roles(organization_id=clients.org_id)
    print_table(items=items, keys=fields, sort_by=sort)


@roles.command()
@click.option("--name", prompt=True)
@click.option("--description", prompt=True)
@click.pass_obj
def create(
    clients: Clients,
    name: str,
    description: str,
) -> None:
    """Create a role"""
    result = clients.contxt_access.post(
        f"{clients.org_id}/roles",
        {
            "name": name,
            "description": description,
        },
    )
    print_item(result)


@roles.command()
@click.argument("target_role")
@click.argument("application_id")
@click.pass_obj
def add_application(clients: Clients, target_role: int, application_id: str) -> None:
    """Add application to role"""
    result = clients.contxt_access.add_role_application(
        organization_id=clients.org_id, target_role_id=target_role, application_id=application_id
    )
    print_item(result)



import time
import webbrowser
from dataclasses import dataclass
from json import dump, load
from pathlib import Path
from typing import Any, Dict, Optional

from auth0.authentication import GetToken
from requests.exceptions import HTTPError

from contxt.services.api import Api

from ..services.auth import AuthService
from ..utils import make_logger
from . import Auth, Token, TokenProvider

logger = make_logger(__name__)


@dataclass
class Env:
    cli_client_id: str
    auth0_tenant_base_url: str


environments = {
    "staging": Env(
        cli_client_id="DMsTIUJINAobrEvgXCpN1koWOGLvdzw6",
        auth0_tenant_base_url="contxt-staging.us.auth0.com",
    ),
    "production": Env(
        cli_client_id="bleED0RUwb7CJ9j7D48tqSiSZRZn29AV", auth0_tenant_base_url="ndustrial.auth0.com"
    ),
}


class DeviceAuthPendingException(Exception):
    pass


class DeviceAuthTimeout(Exception):
    pass


class DeviceAuthDenied(Exception):
    pass


class Auth0DeviceProvider(Api):
    def __init__(self, env: str = "production"):
        self.env = env
        self.base_url = environments[env].auth0_tenant_base_url
        self.auth_service = AuthService(env=env)
        super().__init__(base_url=f"https://{self.base_url}")

    def get_device_code_url(self):
        data = {
            "client_id": environments[self.env].cli_client_id,
            "scope": "offline_access",
            "audience": self.auth_service.client_id,
        }

        return self.post("oauth/device/code", data)

    def get_access_token(self, code_info):
        data = {
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "device_code": code_info["device_code"],
            "client_id": environments[self.env].cli_client_id,
        }

        while True:
            # sleep between polling for the access token for the amount of time
            # recommended in the response
            time.sleep(code_info["interval"])

            try:
                resp = self.post("oauth/token", data)
                return resp
            except HTTPError as e:
                resp = e.response.json()
                if resp["error"] == "authorization_pending":
                    # no authorization yet...we'll wait and try again
                    continue
                elif resp["error"] == "expired_token":
                    raise DeviceAuthTimeout(
                        "The Contxt login timed out. Please run `auth login` command " "to try again."
                    )
                elif resp["error"] == "access_denied":
                    raise DeviceAuthDenied(
                        "You either cancelled the activation or you are currently not "
                        "allowed to use the Contxt CLI. Please contact support if you "
                        "believe this is an error"
                    )
                else:
                    raise e


class UserIdentityProvider(TokenProvider):
    """Concrete `TokenProvider` for a user's identity. The access token is provided
    directly from Auth0 via username and password and granted for offline access, so
    expired tokens are refreshed with a refresh token.

    If `cache_file` is specified, both the `access_token` and `refresh_token`
    are cached there as a JSON blob.
    """

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        audience: str,
        env: str = "production",
        cache_file: Optional[Path] = None,
    ) -> None:
        super().__init__(audience)
        self.env = env
        self.client_id = client_id
        self.client_secret = client_secret
        self.auth_service: GetToken = GetToken(
            environments[env].auth0_tenant_base_url, client_id, client_secret=client_secret
        )
        self._refresh_token: Optional[Token] = None
        self.device_provider = Auth0DeviceProvider(env)

        # Initialize cache
        self._cache_file = cache_file
        self._init_from_cache()

    def _init_from_cache(self) -> None:
        if self._cache_file:
            # Load the cache
            cache = self.read_cache()

            if cache:
                # Token exists in cache, set it
                logger.debug("Setting token from cache")
                self.access_token = cache["access_token"]
                self.refresh_token = cache["refresh_token"]
            else:
                # Token not in cache
                logger.debug("Token not found in cache")

    @TokenProvider.access_token.getter  # type: ignore
    def access_token(self) -> Token:
        """Gets a valid access token for audience `audience`"""
        if self._access_token is None:
            # Token not yet set, fetch one
            logger.debug(f"Fetching new access_token for {self.audience}")
            token_info = self.login()
            self.access_token = token_info["access_token"]
            self.refresh_token = token_info["refresh_token"]
            # Update cache
            self.update_cache()
        elif self._token_expiring():
            # Token expiring soon, refresh it
            logger.debug(f"Refreshing access_token for {self.audience}")
            token_info = self.auth_service.refresh_token(self.refresh_token)
            self.access_token = token_info["access_token"]
            # Update cache
            self.update_cache()
        return self._access_token  # type: ignore

    @property
    def refresh_token(self) -> Token:
        """Gets the refresh token"""
        if self._refresh_token is None:
            # Token not yet set, fetch it now
            self.access_token
        return self._refresh_token  # type: ignore

    @refresh_token.setter
    def refresh_token(self, value: Token) -> None:
        """Sets the refresh token"""
        self._refresh_token = value

    def login(self) -> Dict[str, Any]:
        """Returns an access_token from Auth0 after running the
        "Device Authorization Flow" successfully"""

        # Retrieve a device code to use for the user to login with
        code = self.device_provider.get_device_code_url()

        # Open their default web browser so they can login with their Contxt Account
        print(
            f"You're being redirected to your browser to complete your login. If your browser "
            f"does not automatically open, please navigate to {code['verification_uri_complete']} "
            f"to complete your login."
        )
        print(f"Please ensure the code on the webpage matches the following: {code['user_code']}")
        webbrowser.open(code["verification_uri_complete"])

        # Continuously poll Auth0 to see if they've completed their login yet
        print("Waiting for CLI to be authorized...")
        resp = self.device_provider.get_access_token(code)

        print("Success!")
        return resp

    def reset(self) -> None:
        super().reset()
        self._refresh_token = None

    def read_cache(self) -> Dict[str, Any]:
        if self._cache_file:
            logger.debug(f"Reading cache {self._cache_file}")
            if self._cache_file.is_file():
                with self._cache_file.open("r") as f:
                    return load(f)
        return {}

    def update_cache(self) -> None:
        if self._cache_file:
            logger.debug(f"Updating cache {self._cache_file}")
            cache = {"access_token": self.access_token, "refresh_token": self.refresh_token}
            self._cache_file.parent.mkdir(parents=True, exist_ok=True)
            with self._cache_file.open("w") as f:
                dump(cache, f, indent=4)

    def clear_cache(self) -> None:
        if self._cache_file:
            logger.debug(f"Clearing cache {self._cache_file}")
            if self._cache_file.is_file():
                self._cache_file.unlink()


class UserTokenProvider(TokenProvider):
    """Concrete `TokenProvider` for a user, where `identity_provider` serves as the
    identity provider.
    """

    def __init__(
        self, identity_provider: UserIdentityProvider, audience: str, env: str = "production"
    ) -> None:
        super().__init__(audience)
        self.identity_provider = identity_provider
        self.auth_service = AuthService(env)

    @TokenProvider.access_token.getter  # type: ignore
    def access_token(self) -> Token:
        """Gets a valid access token for audience `audience`"""
        if self._access_token is None or self._token_expiring():
            # Token either not yet set or expiring soon, fetch one
            logger.debug(f"Fetching new access_token for {self.audience}")
            self.access_token = self.auth_service.get_token(
                self.identity_provider.access_token, self.audience
            )["access_token"]
        return self._access_token  # type: ignore


class CliAuth(Auth):
    """Concrete `Auth` for a CLI user, where `identity_provider` authenticates requests
    for an access token for target clients defined by `audience`.

    The access token from Auth0 is cached in a JSON file in a hidden directory,
    meaning that the user will only have to authenticate with their credientials
    once.

    Auth Flow:
    1. User provides Contxt username/password credentials
    2. Retrieve access token directly from Auth0 for our `AuthService`, authenticating
       with above credientials
    3. When authenticating to a service, retrieve access token from our `AuthService`
       for the target service, authenticating with the above Auth0 access token
    """

    def __init__(self, env: str = "production") -> None:
        super().__init__(client_id=environments[env].cli_client_id, client_secret="")
        self.env = env
        self.auth_service = AuthService(env)
        self.identity_provider = UserIdentityProvider(
            env=env,
            client_id=self.client_id,
            client_secret=self.client_secret,
            audience=self.auth_service.client_id,
            cache_file=Path.home() / ".contxt" / "cli_token",
        )

    def get_token_provider(self, audience: str) -> UserTokenProvider:
        """Get `TokenProvider` for audience `audience`"""
        return UserTokenProvider(
            identity_provider=self.identity_provider, audience=audience, env=self.env
        )

    @property
    def user_id(self) -> str:
        return self.identity_provider.decoded_access_token["sub"]

    def get_cluster_config(self, host: str) -> Dict[str, Dict]:
        return self.auth_service.get_cluster_config(
            access_token=self.identity_provider.access_token, host=host
        )

    def logged_in(self) -> bool:
        return bool(self.identity_provider._access_token)

    def login(self) -> None:
        """Force a prompt for the user to login. Note this will happen automatically."""
        if self.logged_in() and not self.query_user("Already logged in. Continue?"):
            return None
        # NOTE: this works by unsetting the current access token, and then
        # calling the getter, which triggers a login prompt
        self.identity_provider.reset()
        self.identity_provider.access_token

    def logout(self) -> None:
        """Logs the user out by clearing the cache"""
        # Reset both the instance and the cache
        self.identity_provider.reset()
        self.identity_provider.clear_cache()

    def query_user(self, question: str) -> bool:
        """Query the user with `question`, and return if user confirmed"""
        choices = ("y", "n")
        # Get only the first character in the response
        answer = input(f"{question} ({'/'.join(choices)}) ").lower()[0:1]
        while answer not in choices:
            answer = input(f"Please enter {' or '.join(choices)}. ").lower()[0:1]
        return answer == "y"



from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from ..auth import Auth
from ..models.contxt import (
    Config,
    ConfigValue,
    Organization,
    OrganizationUser,
    Project,
    Service,
    User,
)
from .api import ApiEnvironment, ConfiguredApi


class ContxtService(ConfiguredApi):
    """Contxt API client"""

    _envs = (
        ApiEnvironment(
            name="production",
            base_url="https://contxt.api.ndustrial.io/v1",
            client_id="8qY2xJob1JAxhmVhIDLCNnGriTM9bct8",
        ),
        ApiEnvironment(
            name="staging",
            base_url="https://contxt.api.staging.ndustrial.io/v1",
            client_id="qGzdTXcmB57zlTp86rYsivG9qEss1lbF",
        ),
    )

    def __init__(self, auth: Auth, env: str = "production", **kwargs) -> None:
        super().__init__(env=env, auth=auth, **kwargs)

    def get_organizations(self) -> List[Organization]:
        resp = self.get("organizations")
        return [Organization.from_api(rec) for rec in resp]

    def get_organization_with_name(self, name: str) -> Optional[Organization]:
        for organization in self.get_organizations():
            if organization.name.lower() == name.lower():
                return organization
        return None

    def create_organization(self, organization: Organization) -> Organization:
        data = organization.post()
        resp = self.post("organizations", data={"name": data["name"], "slug": data["name"]})
        return Organization.from_api(resp)

    def add_user_to_organization(self, user_id: str, organization_id: str) -> OrganizationUser:
        resp = self.post(f"organizations/{organization_id}/users/{user_id}")
        return OrganizationUser.from_api(resp)

    def get_users_for_organization(self, organization_id: str) -> List[User]:
        resp = self.get(f"organizations/{organization_id}/users")
        return [User.from_api(rec) for rec in resp]

    def get_config(self, configuration_id: str) -> Config:
        resp = self.get(f"configurations/{configuration_id}")
        return Config.from_api(resp)

    def get_config_for_client(self, client_id: str, environment_id: str) -> Config:
        params = {"environment_id": environment_id}
        resp = self.get(f"clients/{client_id}/configurations", params=params)
        return Config.from_api(resp)

    def delete_config_value(self, configuration_id: str, value_id: str) -> None:
        # TODO: is this value_id or key?
        self.delete(f"configurations/{configuration_id}/values{value_id}")

    def get_config_values(self, config_id: str, environment_id: str) -> List[ConfigValue]:
        params = {"environment": environment_id}
        resp = self.get(f"configurations/{config_id}/values", params=params)
        return [ConfigValue.from_api(rec) for rec in resp]

    def get_config_values_for_worker(self, worker_id: str, environment_id: str) -> List[ConfigValue]:
        params = {"environment": environment_id}
        resp = self.get(f"workers/{worker_id}/configurations/values", params=params)
        return [ConfigValue.from_api(rec) for rec in resp]

    def start_worker_run(self, client_id: str) -> Dict:
        data = {"start_time": datetime.now()}
        return self.post(f"clients/{client_id}/runs", data=data)

    def end_worker_run(self, client_id: str, run_id: str) -> None:
        data = {"end_time": datetime.now()}
        self.put(f"clients/{client_id}/runs/{run_id}", data=data)

    def create_worker_run_metric(self, run_id: str, key: str, value: Any) -> Dict:
        data = {"key": key, "value": value}
        return self.post(f"runs/{run_id}/metrics", data=data)

    def get_projects(self) -> List[Project]:
        resp = self.get("stacks")
        return [Project.from_api(rec) for rec in resp]

    def get_project(self, project_id) -> Project:
        resp = self.get(f"stacks/{project_id}")
        return Project.from_api(resp)

    def get_services(self, project_id: Union[int, None] = None) -> List[Service]:
        if project_id:
            resp = self.get(f"stacks/{project_id}")
            return [Service.from_api(rec) for rec in resp["Services"]]
        else:
            resp = self.get("services")
            return [Service.from_api(rec) for rec in resp]

    def get_service(self, service_id: int) -> Service:
        resp = self.get(f"services/{service_id}")
        return Service.from_api(resp)

    def create_edge_node_grant(self, edge_node_id: str, to_service_instance_id: int) -> Dict:
        return self.post(
            f"edgenodes/{edge_node_id}/grants", data={"to_service_instance_id": to_service_instance_id}
        )



from datetime import datetime, timedelta
from functools import partial, reduce
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional

import click

from contxt.utils.serializer import Serializer

NOW = datetime.now().replace(microsecond=0)
LAST_WEEK = NOW - timedelta(days=7)

# FIXME
OPTIONAL_PROMPT_KWARGS: Any = {
    "prompt": True,
    "default": "<none>",
    "callback": lambda ctx, param, value: value if value != "<none>" else None,
}


def warn(msg: str):
    click.secho(msg, fg="red", err=True)


def print_table(
    items: Any, keys: Optional[List[str]] = None, sort_by: Optional[str] = None, count: bool = True
) -> None:
    if keys:
        items = pluck(keys=keys, items=items)
        sort_by = sort_by if sort_by in keys else None
    if items:
        print(Serializer.to_table(items, sort_by=sort_by))
    if count and isinstance(items, list):
        print(f"Count: {len(items)}")


def print_item(item: Dict[str, Any]) -> None:
    print_table([{"key": k, "value": v} for k, v in item.items()], count=False)


def pluck(
    keys: List[str],
    items: List[Dict[str, Any]],
    key_sep: str = ".",
    key_func: Callable[[str], str] = lambda k: k,
) -> List[Dict[str, Any]]:
    """Pluck `keys` from `items`. Nested keys can be specified by concatenating with
    `key_sep` (i.e. `key1.key2`). Returned keys are set via `key_func`."""
    return [
        {
            key_func(k): reduce(
                lambda d, k: d[k] if isinstance(d, dict) else getattr(d, k), k.split(key_sep), d
            )
            for k in keys
        }
        for d in items
    ]


def csv_callback(options: List[str], all: Optional[List[str]] = None):
    def _callback(ctx, param, value):
        # Parse
        if isinstance(value, str) and value == "all":
            fields = all or options
        elif isinstance(value, str):
            fields = [f.strip() for f in value.split(",")]
        else:
            fields = value
        # Validate
        for f in fields:
            if f not in options:
                raise click.BadParameter(f"'{f}' is not valid. Choose from {', '.join(options)}.")
        return fields

    return _callback


def fields_option(func: Optional[click.Command] = None, **kwargs) -> click.Command:
    if func is None:
        return partial(fields_option, **kwargs)  # type: ignore
    return click.option(
        "--fields",
        callback=csv_callback(options=[f.attr_key for f in kwargs.pop("obj")._api_fields], all=[]),
        help="Comma-delimited list of fields to return",
        **kwargs,
    )(func)


def sort_option(func: Optional[click.Command] = None, **kwargs) -> click.Command:
    if func is None:
        return partial(sort_option, **kwargs)  # type: ignore
    return click.option("--sort", help="Field to sort by", **kwargs)(func)


# https://github.com/pallets/click/issues/405
class ClickPath(click.Path):
    def convert(self, *args, **kwargs) -> Any:
        return Path(super().convert(*args, **kwargs))


class Date(click.DateTime):
    def __init__(self) -> None:
        super().__init__(formats=["%Y-%m-%d"])

    def convert(self, *args, **kwargs) -> Any:
        return super().convert(*args, **kwargs).date()


def str_to_bool(s):
    if s.lower() in ("true", "t", "1"):
        return True
    elif s.lower() in ("false", "f", "0"):
        return False
    else:
        raise ValueError("Cannot covert {} to a bool".format(s))



from typing import List

import click

from contxt.cli.clients import Clients
from contxt.cli.utils import fields_option, print_item, print_table, sort_option
from contxt.models.contxt import Service


@click.group()
def services() -> None:
    """Services."""


@services.command()
@click.argument("id", default="")  # HACK: make an optional argument
@click.pass_obj
@fields_option(default="id, slug, project_id, description", obj=Service)
@sort_option(default="slug")
def get(clients: Clients, id: str, fields: List[str], sort: str) -> None:
    """Get service(s)"""
    items = (
        [clients.contxt_deployments.get(f"{clients.org_id}/services/{id}")]
        if id
        else clients.contxt_deployments.get(f"{clients.org_id}/services")
    )
    print_table(items=items, keys=fields, sort_by=sort)


@services.command()
@click.option("--project-id", prompt=True)
@click.option("--slug", prompt=True)
@click.option("--type", type=click.Choice(["API", "Application", "Worker"]), prompt=True)
@click.option(
    "--deployment-strategy",
    type=click.Choice(["self-managed", "contxt-managed", "unmanaged"]),
    prompt=True,
)
@click.pass_obj
def create(clients: Clients, project_id: str, slug: str, type: str, deployment_strategy: str) -> None:
    """Create service"""
    result = clients.contxt_deployments.post(
        f"{clients.org_id}/services",
        json={
            "project_id": project_id,
            "slug": slug,
            "type": type,
            "deployment_strategy": deployment_strategy,
        },
    )
    print_item(result)


@services.command()
@click.argument("id")
@click.pass_obj
def delete(clients: Clients, id: str) -> None:
    """Delete service"""
    clients.contxt_deployments.delete(f"{clients.org_id}/services/{id}")



from dataclasses import dataclass
from pprint import pformat
from typing import Any, ClassVar, List, Optional

from ..utils.serializer import Serializer
from . import ApiField, ApiObject


def pretty_print(obj: Any) -> str:
    d = Serializer.to_dict(obj)
    return pformat(d, indent=4)


@dataclass
class PublisherStats(ApiObject):
    _api_fields: ClassVar = (
        ApiField("msgRateIn", attr_key="msg_rate_in"),
        ApiField("producerId", attr_key="producer_id"),
        ApiField("connectedSince", attr_key="connected_since"),
    )

    msg_rate_in: str
    producer_id: str
    connected_since: str

    def __str__(self) -> str:
        return pretty_print(self)


@dataclass
class ConsumerStats(ApiObject):
    _api_fields: ClassVar = (
        ApiField("msgRateOut", attr_key="msg_rate_out", data_type=float),
        ApiField("msgRateRedeliver", attr_key="msg_rate_redeliver", data_type=float),
        ApiField("unackedMessages", attr_key="unacked_messages", data_type=int),
        ApiField(
            "blockedConsumerOnUnackedMsgs", attr_key="blocked_consumer_on_unacked_msgs", data_type=bool
        ),
        ApiField("connectedSince", attr_key="connected_since"),
    )

    msg_rate_out: float
    msg_rate_redeliver: float
    unacked_messages: int
    blocked_consumer_on_unacked_msgs: bool
    connected_since: str

    def __str__(self) -> str:
        return pretty_print(self)


@dataclass
class SubscriberStats(ApiObject):
    _api_fields: ClassVar = (
        ApiField("name", optional=True),
        ApiField("msgRateOut", attr_key="msg_rate_out", data_type=float),
        ApiField("msgRateRedeliver", attr_key="msg_rate_redeliver", data_type=float),
        ApiField("msgBacklog", attr_key="msg_backlog", data_type=int),
        ApiField(
            "blockedSubscriptionOnUnackedMsgs",
            attr_key="blocked_subscription_on_unacked_msgs",
            data_type=bool,
        ),
        ApiField("unackedMessages", attr_key="unacked_messages", data_type=int),
        ApiField("consumers", data_type=ConsumerStats),
    )

    msg_rate_out: float
    msg_rate_redeliver: float
    msg_backlog: int
    blocked_subscription_on_unacked_msgs: bool
    unacked_messages: int
    consumers: List[ConsumerStats]
    name: Optional[str] = None

    def __str__(self) -> str:
        return pretty_print(self)


@dataclass
class Channel(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id", data_type=str, optional=True),
        ApiField("name", data_type=str),
        ApiField("organization_id", data_type=str),
        ApiField("service_id", data_type=str),
    )

    name: str
    service_id: str
    organization_id: str
    id: Optional[str] = None

    def __str__(self) -> str:
        return pretty_print(self)


@dataclass
class ChannelStats(ApiObject):
    _api_fields: ClassVar = (
        ApiField("bytesInCounter", attr_key="bytes_in_counter", data_type=int),
        ApiField("msgInCounter", attr_key="msg_in_counter", data_type=int),
        ApiField("bytesOutCounter", attr_key="bytes_out_counter", data_type=int),
        ApiField("msgOutCounter", attr_key="msg_out_counter", data_type=int),
        ApiField("backlogSize", attr_key="backlog_size", data_type=int),
        ApiField("msgRateIn", attr_key="msg_rate_in", data_type=float),
        ApiField("msgRateOut", attr_key="msg_rate_out", data_type=float),
        ApiField("publishers", data_type=PublisherStats),
        ApiField(
            "subscriptions",
            data_type=lambda o: [SubscriberStats.from_api({**v, "name": k}) for k, v in o.items()],
        ),
    )

    bytes_in_counter: int
    msg_in_counter: int
    bytes_out_counter: int
    msg_out_counter: int
    backlog_size: int
    msg_rate_in: float
    msg_rate_out: float
    publishers: List[PublisherStats]
    subscriptions: List[SubscriberStats]

    def __str__(self) -> str:
        return pretty_print(self)



import json
from dataclasses import dataclass, field
from typing import Any, Dict, List, Union

from jwt import PyJWTError, algorithms, decode, get_unverified_header

from ..services import AuthService
from ..utils import make_logger

logger = make_logger(__name__)


class InvalidTokenError(Exception):
    """Invalid JWT"""


@dataclass
class TokenValidator:
    audience: str
    issuer: str
    public_key: Union[str, Dict]
    algorithms: List[str] = field(default_factory=lambda: ["RS256"])

    def _get_public_key(self, token: str) -> str:
        if isinstance(self.public_key, dict):
            header = get_unverified_header(token)
            return self.public_key[header["kid"]]
        return self.public_key

    def validate(self, token: str, **kwargs) -> Dict[str, Any]:
        """Validates `token`"""
        try:
            return decode(
                jwt=token,
                key=self._get_public_key(token),
                algorithms=self.algorithms,
                audience=self.audience,
                issuer=self.issuer,
                **kwargs,
            )
        except PyJWTError as e:
            logger.exception(e)
            raise InvalidTokenError(f"INVALID_TOKEN: {str(e)}")
        except Exception as e:
            logger.exception(e)
            raise InvalidTokenError("MALFORMED_TOKEN: failed to parse token")


class ContxtTokenValidator(TokenValidator):
    def __init__(self, audience: str) -> None:
        api = AuthService()
        super().__init__(
            audience=audience,
            issuer=api.base_url,
            public_key={
                jwk["kid"]: algorithms.RSAAlgorithm.from_jwk(json.dumps(jwk))
                for jwk in api.get_jwks()["keys"]
            },
            algorithms=["RS256"],
        )



from typing import List

from requests.exceptions import HTTPError

from ..auth import Auth
from ..models.contxt import Cluster, EdgeNode, EdgeNodeGrant, ServiceInstanceGrant
from .api import ApiEnvironment, ConfiguredApi


class ContxtDeploymentService(ConfiguredApi):
    """Contxt API client"""

    _envs = (
        ApiEnvironment(
            name="production",
            base_url="https://contxt.api.ndustrial.io/deploy/v1",
            client_id="8qY2xJob1JAxhmVhIDLCNnGriTM9bct8",
        ),
        ApiEnvironment(
            name="staging",
            base_url="https://contxt.api.staging.ndustrial.io/deploy/v1",
            client_id="qGzdTXcmB57zlTp86rYsivG9qEss1lbF",
        ),
    )

    def __init__(self, auth: Auth, env: str = "production", **kwargs) -> None:
        super().__init__(env=env, auth=auth, **kwargs)

    def get_clusters(self, organization_id: str) -> List[Cluster]:
        return [Cluster.from_api(rec) for rec in self.get(f"{organization_id}/clusters")]

    def get_cluster(self, organization_id: str, cluster_slug: str) -> Cluster:
        resp = self.get(f"{organization_id}/clusters/{cluster_slug}")
        return Cluster.from_api(resp)

    def register_cluster(self, organization_id: str, cluster: Cluster) -> None:
        obj = cluster.post()
        try:
            resp = self.post(f"{organization_id}/clusters", json=obj)
            return Cluster.from_api(resp)
        except HTTPError:
            pass
            return

    def get_edge_nodes(self, organization_id: str, project_slug: int) -> List[EdgeNode]:
        resp = self.get(f"{organization_id}/projects/{project_slug}/edgenodes")
        return [EdgeNode.from_api(rec) for rec in resp]

    def get_edge_node_grants(self, organization_id: str, edge_node_id: int) -> List[EdgeNodeGrant]:
        resp = self.get(f"{organization_id}/edgenodes/{edge_node_id}/grants")
        return [EdgeNodeGrant.from_api(rec) for rec in resp]

    def get_service_instance_grants(
        self, organization_id: str, service_instance_id: int
    ) -> List[ServiceInstanceGrant]:
        resp = self.get(f"{organization_id}/service_instances/{service_instance_id}/grants")
        return [ServiceInstanceGrant.from_api(rec) for rec in resp]



from typing import Dict, List

from ..auth import Auth
from ..models.contxt import Role
from .api import ApiEnvironment, ConfiguredApi


class ContxtAccessService(ConfiguredApi):
    """Contxt API client"""

    _envs = (
        ApiEnvironment(
            name="production",
            base_url="https://contxt.api.ndustrial.io/access/v1",
            client_id="8qY2xJob1JAxhmVhIDLCNnGriTM9bct8",
        ),
        ApiEnvironment(
            name="staging",
            base_url="https://contxt.api.staging.ndustrial.io/access/v1",
            client_id="qGzdTXcmB57zlTp86rYsivG9qEss1lbF",
        ),
    )

    def __init__(self, auth: Auth, env: str = "production", **kwargs) -> None:
        super().__init__(env=env, auth=auth, **kwargs)

    def get_roles(self, organization_id: str) -> List[Role]:
        return [Role.from_api(rec) for rec in self.get(f"{organization_id}/roles")]

    def add_role_application(
        self, organization_id: str, target_role_id: str, application_id: int
    ) -> Dict:
        return self.post(f"{organization_id}/roles/{target_role_id}/applications/{application_id}")



"""Data models for API clients"""

from abc import ABC, abstractmethod
from ast import literal_eval
from copy import deepcopy
from datetime import date as _date
from datetime import datetime as _datetime
from datetime import timedelta, timezone
from importlib import import_module
from typing import Any, Callable, Dict, Optional, Tuple, Union

from ..utils import make_logger
from ..utils.serializer import Serializer

logger = make_logger(__name__)


class Parsers:
    """Parsers to deserialize JSON as Python"""

    boolean = bool
    number = float
    string = str

    @staticmethod
    def date(datestamp: str) -> _date:
        return _date.fromisoformat(datestamp[:10])

    @staticmethod
    def datetime(timestamp: str) -> _datetime:
        return _datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S.%fZ").replace(tzinfo=timezone.utc)

    @staticmethod
    def unknown(value: Any) -> Any:
        # First, try a general parser (supports strings, numbers, tuples,
        # lists, dicts, booleans (True/False strings only), and None)
        try:
            return literal_eval(value)
        except (SyntaxError, ValueError):
            pass
        # Next, fall back to a datetime parser
        try:
            return Parsers.datetime(value)
        except (TypeError, ValueError):
            pass
        # Next, convert bool-type strings
        try:
            if value.lower() in ("yes", "true"):
                return True
            elif value.lower() in ("no", "false"):
                return False
        except (AttributeError, TypeError, ValueError):
            pass
        # Failed, return original value
        return value


class Formatters:
    """Formatters to serialize Python as JSON"""

    @staticmethod
    def date(d: _date) -> str:
        return d.isoformat()

    @staticmethod
    def datetime(dt: _datetime, strict: bool = True) -> str:
        if dt.utcoffset() != timedelta() and strict:
            # Require timezone to be UTC
            raise AssertionError(f"Datetime must be UTC, not {dt.tzinfo}")
        # NOTE: almost exactly the same as isoformat(), but ensures
        # microseconds are always represented
        return dt.astimezone(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%fZ")

    @staticmethod
    def normalize_label(text: str) -> str:
        return text.lower().replace(" ", "_")


# TODO: Need a way to track changed attributes
# TODO: This custom schema-validation can be replaced by a more robust
# (although slower) implementation: see https://github.com/marshmallow-code/marshmallow
class ApiField:
    """A field retrieved from an API service.

    `api_key` is the expected key, `attr_key` is the desired class attribute key,
    `data_type` is the object type.
    """

    def __init__(
        self,
        api_key: str,
        attr_key: Optional[str] = None,
        data_type: Union[Callable, str] = str,
        creatable: bool = False,
        updatable: bool = False,
        optional: bool = False,
    ) -> None:
        self.api_key = api_key
        self.attr_key = attr_key or api_key
        self._data_type = data_type
        self.creatable = creatable
        self.updatable = updatable
        self.optional = optional

    @property
    def data_type(self) -> Callable[[Any], Any]:
        if isinstance(self._data_type, str):
            # Load callable from str
            # NOTE: this is to delay the type assignment to instance creation,
            # as the type might not yet be defined at class creation
            modname, qualname_separator, qualname = self._data_type.partition(":")
            obj = import_module(modname)
            if qualname_separator:
                for attr in qualname.split("."):
                    obj = getattr(obj, attr)
            self._data_type = obj  # type: ignore
        return self._data_type  # type: ignore


class ApiObject(ABC):
    """An abstract base class for a response from an API. This class serves to
    take a raw response from an API and create a parsed Python object.
    """

    def __str__(self) -> str:
        return Serializer.to_table(self)

    @property  # type: ignore
    @classmethod
    @abstractmethod
    def _api_fields(cls) -> Tuple[ApiField, ...]:
        pass

    @property
    def creatable_fields(self) -> Dict[str, ApiField]:
        cls = type(self)
        if not hasattr(cls, "_creatable_fields"):
            cls._creatable_fields = {  # type: ignore
                f.attr_key: f for f in cls._api_fields if f.creatable  # type: ignore
            }
        return cls._creatable_fields  # type: ignore

    @property
    def updatable_fields(self) -> Dict[str, ApiField]:
        cls = type(self)
        if not hasattr(cls, "_updatable_fields"):
            cls._updatable_fields = {  # type: ignore
                f.attr_key: f for f in cls._api_fields if f.updatable  # type: ignore
            }
        return cls._updatable_fields  # type: ignore

    @classmethod
    def from_api(cls, api_dict: Dict) -> Any:
        api_dict = deepcopy(api_dict)
        # Create clean dictionary to pass to init
        clean_dict = {}
        for field in cls._api_fields:  # type: ignore
            if field.api_key not in api_dict and not field.optional:
                m = cls.__module__
                n = cls.__name__
                raise KeyError(
                    f"{m}.{n} : Required API field '{field.api_key}' is missing from"
                    f" response: {api_dict}"
                )
            clean_dict[field.attr_key] = cls.clean_api_value(
                field=field, value=api_dict.pop(field.api_key, None)
            )

        # NOTE: unexpected keys: api_dict.keys()
        return cls(**clean_dict)  # type: ignore

    @classmethod
    def clean_api_value(cls, field: ApiField, value: Any) -> Any:
        if value is None:
            # No value
            return value
        elif isinstance(value, (list, tuple)):
            # Value is a list, clean each item
            return [cls.clean_api_value(field, v) for v in value]
        # elif issubclass(field.data_type, ApiObject):
        elif callable(getattr(field.data_type, "from_api", None)):
            # Type is an ApiObject, apply from_api instead of init
            return field.data_type.from_api(value)  # type: ignore
        else:
            # Apply type
            return field.data_type(value)

    def post(self) -> Dict[str, Any]:
        """Gets data for a POST request"""
        # Transform api fields to dict
        d = Serializer.to_dict(self, key_filter=lambda k: k in set(self.creatable_fields.keys()))
        # Swap attr_keys for api_keys
        return {self.creatable_fields[k].api_key: v for k, v in d.items()}

    def put(self) -> Dict[str, Any]:
        """Gets data for a PUT request"""
        # Transform api fields to dict
        d = Serializer.to_dict(self, key_filter=lambda k: k in set(self.updatable_fields.keys()))
        # Swap attr_keys for api_keys
        return {self.updatable_fields[k].api_key: v for k, v in d.items()}



"""General utilities"""

from datetime import datetime
from functools import partial, wraps
from logging import Logger, getLogger
from time import time
from typing import Callable, Dict, Optional


def make_logger(name: str, level: Optional[str] = None) -> Logger:
    """Get logger with name `name` and level `level`"""
    logger = getLogger(name)
    if level:
        logger.setLevel(level.upper())
    return logger


# Ref: https://docs.python.org/3/library/datetime.html#determining-if-an-object-is-aware-or-naive
def is_datetime_aware(dt: datetime) -> bool:
    """Returns if datetime `dt` is timezone-aware"""
    return dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None


def timed(func: Optional[Callable] = None, *, printer: Callable[[str], None] = print) -> Callable:
    """Decorator to print `func`'s runtime, via `printer`"""

    if func is None:
        return partial(timed, printer=printer)

    @wraps(func)
    def wrapper(*args, **kwargs):
        t0 = time()
        res = func(*args, **kwargs)
        printer(f"{func.__name__}'s elapsed time: {time() - t0} s")
        return res

    return wrapper


# Source: https://boltons.readthedocs.io/en/latest/_modules/boltons/cacheutils.html#cachedproperty
class cachedproperty:
    """The ``cachedproperty`` is used similar to :class:`property`, except
    that the wrapped method is only called once. This is commonly used
    to implement lazy attributes.

    After the property has been accessed, the value is stored on the
    instance itself, using the same name as the cachedproperty. This
    allows the cache to be cleared with :func:`delattr`, or through
    manipulating the object's ``__dict__``.
    """

    def __init__(self, func):
        self.__doc__ = getattr(func, "__doc__")
        self.__isabstractmethod__ = getattr(func, "__isabstractmethod__", False)
        self.func = func

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value

    def __repr__(self):
        return f"<{self.__class__.__name__} func={self.func}>"



import sgqlc.types
import sgqlc.operation
from . import nionic_schema

_schema = nionic_schema
_schema_root = _schema.nionic_schema

__all__ = ('Operations',)


def mutation_create_facility():
    _op = sgqlc.operation.Operation(_schema_root.mutation_type, name='createFacility', variables=dict(name=sgqlc.types.Arg(sgqlc.types.non_null(_schema.String)), slug=sgqlc.types.Arg(sgqlc.types.non_null(_schema.String)), timezoneName=sgqlc.types.Arg(sgqlc.types.non_null(_schema.String))))
    _op_create_facility = _op.create_facility(input={'facility': {'name': sgqlc.types.Variable('name'), 'slug': sgqlc.types.Variable('slug'), 'timezoneName': sgqlc.types.Variable('timezoneName')}})
    _op_create_facility_facility = _op_create_facility.facility()
    _op_create_facility_facility.id()
    return _op


class Mutation:
    create_facility = mutation_create_facility()


def query_facilities():
    _op = sgqlc.operation.Operation(_schema_root.query_type, name='facilities')
    _op_facilities = _op.facilities()
    _op_facilities_nodes = _op_facilities.nodes()
    _op_facilities_nodes.id()
    _op_facilities_nodes.name()
    _op_facilities_nodes.slug()
    _op_facilities_nodes.address()
    _op_facilities_nodes.city()
    _op_facilities_nodes.state()
    _op_facilities_nodes.zip()
    _op_facilities_nodes.timezone_name()
    return _op


def query_metric_labels():
    _op = sgqlc.operation.Operation(_schema_root.query_type, name='metricLabels')
    _op_source_labels = _op.source_labels()
    _op_source_labels.source_id()
    _op_source_labels.label()
    return _op


def query_main_services():
    _op = sgqlc.operation.Operation(_schema_root.query_type, name='mainServices', variables=dict(facilityId=sgqlc.types.Arg(sgqlc.types.non_null(_schema.Int))))
    _op_main_services = _op.main_services(filter={'facilityId': {'equalTo': sgqlc.types.Variable('facilityId')}})
    _op_main_services_nodes = _op_main_services.nodes()
    _op_main_services_nodes.id()
    _op_main_services_nodes.facility_id()
    _op_main_services_nodes.name()
    _op_main_services_nodes.type()
    _op_main_services_nodes_usage = _op_main_services_nodes.usage()
    _op_main_services_nodes_usage.id()
    _op_main_services_nodes_usage.data_source_name()
    _op_main_services_nodes_usage.name()
    _op_main_services_nodes_usage.alias()
    _op_main_services_nodes_usage.data_type()
    _op_main_services_nodes_demand = _op_main_services_nodes.demand()
    _op_main_services_nodes_demand.id()
    _op_main_services_nodes_demand.data_source_name()
    _op_main_services_nodes_demand.name()
    _op_main_services_nodes_demand.alias()
    _op_main_services_nodes_demand.data_type()
    _op_main_services_nodes.created_at()
    _op_main_services_nodes.updated_at()
    return _op


class Query:
    facilities = query_facilities()
    main_services = query_main_services()
    metric_labels = query_metric_labels()


class Operations:
    mutation = Mutation
    query = Query



from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from json import loads
from typing import Any, ClassVar, Dict, List, Optional

from requests import Request

from . import ApiField, ApiObject, Parsers
from .events import Owner


class Window(Enum):
    RAW = 0
    MINUTELY = 60
    QUARTER_HOURLY = 900
    HOURLY = 3600


class FieldValueType(Enum):
    BOOLEAN = "boolean"
    NUMERIC = "numeric"
    STRING = "string"


@dataclass
class Field(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id", data_type=int),
        ApiField("field_name", attr_key="name", optional=True),
        ApiField("label", creatable=True, updatable=True),
        ApiField("output_id", data_type=int, creatable=True),
        ApiField("field_descriptor", creatable=True),
        ApiField("field_human_name"),
        ApiField("units", creatable=True, updatable=True),
        ApiField("scalar", data_type=float, optional=True, creatable=True, updatable=True),
        ApiField("divisor", data_type=float, optional=True, creatable=True, updatable=True),
        ApiField("value_type", data_type=FieldValueType, optional=True, creatable=True),
        ApiField("feed_key", optional=True, creatable=True),
        ApiField("FieldGroupingField", attr_key="field_grouping_field", data_type=dict, optional=True),
        ApiField("is_hidden", data_type=bool, creatable=True, updatable=True, optional=True),
        ApiField("is_default", data_type=bool, optional=True, creatable=True, updatable=True),
        ApiField("is_totalizer", data_type=bool, optional=True, creatable=True, updatable=True),
        ApiField("can_aggregate", data_type=bool, optional=True),
        ApiField("is_windowed", data_type=bool, optional=True),
        ApiField("status", optional=True),
        ApiField("created_at", data_type=Parsers.datetime, optional=True),
        ApiField("updated_at", data_type=Parsers.datetime, optional=True),
    )

    label: str
    field_descriptor: str
    units: str
    id: Optional[int] = None
    output_id: Optional[str] = None
    name: Optional[str] = None
    field_human_name: Optional[str] = None
    feed_key: Optional[str] = None
    field_grouping_field: Optional[Dict] = None
    value_type: Optional[FieldValueType] = None
    scalar: Optional[float] = None
    divisor: Optional[float] = None
    is_hidden: Optional[bool] = None
    is_totalizer: Optional[bool] = None
    is_windowed: Optional[bool] = None
    is_default: Optional[bool] = None
    can_aggregate: Optional[bool] = None
    status: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


@dataclass
class FieldCategory(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("name"),
        ApiField("description"),
        ApiField("organization_id"),
        ApiField("parent_category_id"),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
        ApiField(
            "FieldGroupings",
            attr_key="field_groupings",
            data_type=lambda v: FieldGrouping.from_api(v),
            optional=True,
        ),
        ApiField(
            "Subcategories",
            attr_key="subcategories",
            data_type=lambda v: FieldCategory.from_api(v),
            optional=True,
        ),
    )

    id: str
    name: str
    description: str
    organization_id: str
    parent_category_id: str
    created_at: datetime
    updated_at: datetime
    field_groupings: Optional[List["FieldGrouping"]] = None
    subcategories: Optional[List["FieldCategory"]] = None


@dataclass
class FieldGrouping(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("label"),
        ApiField("slug"),
        ApiField("description"),
        ApiField("facility_id", data_type=int),
        ApiField("owner_id"),
        ApiField("is_public", data_type=bool),
        ApiField("field_category_id"),
        ApiField("Owner", attr_key="owner", data_type=Owner, optional=True),
        ApiField("FieldCategory", attr_key="field_category", data_type=FieldCategory, optional=True),
        ApiField("Fields", attr_key="fields", data_type=Field, optional=True),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    id: str
    label: str
    slug: str
    description: str
    facility_id: int
    is_public: bool
    owner_id: str
    field_category_id: str
    created_at: datetime
    updated_at: datetime
    owner: Optional[Owner] = None
    field_category: Optional[FieldCategory] = None
    fields: List[Field] = field(default_factory=list)


@dataclass
class UnprovisionedField(ApiObject):
    _api_fields: ClassVar = (ApiField("field_descriptor"), ApiField("assumed_type"))

    field_descriptor: str
    assumed_type: str


@dataclass
class Feed(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id", data_type=int),
        ApiField("feed_type_id", data_type=int),
        ApiField("down_after"),
        ApiField("key"),
        ApiField("facility_id", data_type=int),
        ApiField("timezone"),
        ApiField("token"),
        ApiField("status"),
        ApiField("critical_threshold", data_type=float),
        ApiField("status_event_id"),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
        ApiField("routing_keys", data_type=lambda o: loads(o)),
        ApiField("is_paused", data_type=bool),
        ApiField("owner_id"),
        ApiField("owner", data_type=Owner),
        ApiField("feed_type", data_type=dict),
        ApiField("Metrics", attr_key="metrics", data_type=dict, optional=True),
    )

    id: int
    facility_id: int
    feed_type_id: int
    down_after: str
    key: str
    timezone: str
    token: str
    status: str
    critical_threshold: float
    status_event_id: str
    created_at: datetime
    updated_at: datetime
    routing_keys: str
    is_paused: bool
    owner_id: str
    feed_type: str
    owner: Owner
    metrics: Optional[Dict] = None


@dataclass
class FieldTimeSeries:
    field: Field
    time_series: Dict[datetime, Any]


@dataclass
class BatchRequest:
    method: str
    uri: str
    body: Optional[str] = None

    @staticmethod
    def from_request(request: Request) -> "BatchRequest":
        # NOTE: this handles url-encoding parameters and other low-level translations
        r = request.prepare()
        return BatchRequest(method=r.method, uri=r.url, body=r.body)  # type: ignore

    def to_api(self) -> Dict[str, str]:
        d = {"method": self.method, "uri": self.uri}
        if self.body:
            d["body"] = self.body
        return d


BatchRequests = Dict[str, BatchRequest]


@dataclass
class BatchResponse:
    body: Any
    headers: Dict[str, Any]
    statusCode: int

    @property
    def ok(self) -> bool:
        return self.statusCode == 200


BatchResponses = Dict[str, BatchResponse]



from datetime import date, datetime, timedelta, timezone
from typing import Iterable

from ..auth import Auth
from ..models.ems import ResourceType, UtilityContract, UtilitySpend, UtilityUsage
from .api import ApiEnvironment, ConfiguredApi
from .pagination import PagedRecords


class EmsService(ConfiguredApi):
    """EMS API client"""

    _envs = (
        ApiEnvironment(
            name="production",
            base_url="https://{tenant}.api.ndustrial.io/rest/ems/v1",
            client_id="e2IT0Zm9RgGlDBkLa2ruEcN9Iop6dJAS",
        ),
        ApiEnvironment(
            name="staging",
            base_url="https://{tenant}.api.staging.ndustrial.io/rest/ems/v1",
            client_id="vMV67yaRFgjBB1JFbT3vXBOlohFdG1I4",
        ),
    )

    def __init__(
        self, auth: Auth, org_id: str, org_slug: str, env: str = "production", **kwargs
    ) -> None:
        super().__init__(env=env, auth=auth, **kwargs)
        self.org_id = org_id
        self.base_url = self.base_url.format(tenant=org_slug)

    def get_monthly_utility_spend(
        self,
        facility_id: int,
        resource_type: ResourceType = ResourceType.ELECTRIC,
        start_date: date = date.today() - timedelta(days=3600),
        end_date: date = date.today(),
        pro_forma: bool = False,
        exclude_account_charges: bool = False,
    ) -> UtilitySpend:
        """Get monthly utility spend for facility `facility_id` and resource
        `resource_type` (for now, this must be "electric" but will be expanded).
        Note `start_date` defaults to 10 years ago (the API's maximum limit) and
        `end_date` defaults to today."""
        resp = self.get(
            f"{self.org_id}/facilities/{facility_id}/utility/spend/monthly",
            params={
                "type": resource_type.value,
                "date_start": start_date.strftime("%Y-%m"),
                "date_end": end_date.strftime("%Y-%m"),
                "pro_forma": pro_forma,
                "exclude_account_charges": exclude_account_charges,
            },
        )
        return UtilitySpend.from_api(resp)

    def get_monthly_utility_usage(
        self,
        facility_id: int,
        resource_type: ResourceType = ResourceType.ELECTRIC,
        start_date: date = date.today() - timedelta(days=3600),
        end_date: date = date.today(),
        pro_forma: bool = False,
    ) -> UtilityUsage:
        """Get monthly utility usage for facility `facility_id` and resource
        `resource_type` (for now, this must be "electric" but will be expanded).
        Note `start_date` defaults to 10 years ago (the API's maximum limit) and
        `end_date` defaults to today."""
        resp = self.get(
            f"{self.org_id}/facilities/{facility_id}/utility/usage/monthly",
            params={
                "type": resource_type.value,
                "date_start": start_date.strftime("%Y-%m"),
                "date_end": end_date.strftime("%Y-%m"),
                "pro_forma": pro_forma,
            },
        )
        return UtilityUsage.from_api(resp)

    def get_usage(
        self,
        facility_id: int,
        interval: str,
        resource_type: ResourceType = ResourceType.ELECTRIC,
        start: datetime = datetime.now(timezone.utc) - timedelta(days=365),
        end: datetime = datetime.now(timezone.utc),
    ) -> UtilityUsage:
        """Get utility usage for facility `facility_id` and resource `resource_type`"""
        resp = self.get(
            f"{self.org_id}/facilities/{facility_id}/usage/{interval}",
            params={
                "type": resource_type.value,
                "time_start": int(start.timestamp()),
                "time_end": int(end.timestamp()),
            },
        )
        return UtilityUsage.from_api(resp)

    def get_utility_contracts_for_facility(self, facility_id: int) -> Iterable[UtilityContract]:
        return PagedRecords(
            api=self,
            url=f"{self.org_id}/facilities/{facility_id}/contracts",
            record_parser=UtilityContract.from_api,
        )



import click

from contxt.cli.clients import Clients


@click.group()
def auth() -> None:
    """Authenticate with Contxt."""


@auth.command()
@click.pass_obj
def login(client: Clients) -> None:
    """Login to Contxt."""
    client.auth.login()


@auth.command()
@click.pass_obj
def logout(client: Clients) -> None:
    """Logout of Contxt."""
    client.auth.logout()


@auth.command()
@click.argument("audience")
@click.pass_obj
def token(client: Clients, audience: str) -> None:
    """Fetch an access token for an API."""
    token = client.auth.get_token_provider(audience).access_token
    print(token)



import logging
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import ClassVar, Optional

import requests

from . import ApiField, ApiObject, Parsers

logger = logging.getLogger(__name__)


@dataclass
class FileRead(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("file_id"),
        ApiField("temporary_url"),
        ApiField("expires_at"),
        ApiField("action"),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
        ApiField("file_created_at", data_type=Parsers.datetime),
    )

    id: int
    file_id: int
    temporary_url: str
    expires_at: datetime
    action: str
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    file_created_at: Optional[datetime] = None

    def download(self, path: Path, attempts: int = 1) -> None:
        path.parent.mkdir(parents=True, exist_ok=True)
        attempt = 1
        while attempt <= attempts:
            try:
                r = requests.get(self.temporary_url)
                with path.open("wb") as f:
                    f.write(r.content)
                return
            except requests.exceptions.HTTPError as e:
                logger.warning(f"Exception raised during download {e}")
                attempt += 1
        raise RuntimeError(f"Failed to download file from {self.temporary_url}")



from dataclasses import dataclass
from datetime import datetime
from math import ceil
from typing import Any, Callable, Dict, Generic, Iterator, List, Optional, Tuple, TypeVar, Union

from contxt.services.api import BaseGraphService

from ..models import Parsers
from ..utils.object_mapper import ObjectMapper
from .api import Api

T = TypeVar("T")
Record = Dict[str, Any]
DataPoint = Tuple[datetime, Any]


@dataclass
class PageOptions:
    per_page: int = 1000
    order_by: Optional[str] = None  # NOTE: ideally this should be an enum, if possible
    reverse_order: Optional[bool] = None

    def __post_init__(self) -> None:
        assert self.per_page > 0, f"per_page must be a positive integer, not {self.per_page}"

    def to_api(self, index: int) -> Dict:
        return {
            "offset": self.per_page * index,
            "limit": self.per_page,
            "orderBy": self.order_by,
            "reverseOrder": self.reverse_order,
        }


@dataclass
class PageMetadata:
    totalRecords: int
    offset: int = 0


@dataclass
class Page:
    records: List[Record]
    _metadata: PageMetadata

    def __len__(self) -> int:
        return len(self.records)

    def __iter__(self) -> Iterator[Record]:
        yield from self.records

    def __getitem__(self, index: Union[int, slice]) -> Union[Record, List[Record]]:
        return self.records[index]


class PagedRecords(Generic[T]):
    def __init__(
        self,
        api: Api,
        url: str,
        params: Optional[Dict] = None,
        options: Optional[PageOptions] = None,
        record_parser: Optional[Callable[[Record], T]] = None,
    ):
        self.api = api
        self.url = url
        self.params = params or {}
        self.options = options or PageOptions()
        self.record_parser = record_parser or (lambda x: x)  # type: ignore

        # Fetch first page
        self.page_index = 0
        self.page = self.get_page(index=self.page_index, force=True)

    def __len__(self) -> int:
        return self.total_records

    def __iter__(self) -> Iterator[T]:
        for page_index in range(self.total_pages):
            yield from self.get_page(page_index)  # type: ignore

    def __getitem__(self, index: Union[int, slice]) -> Union[T, List[T]]:
        if isinstance(index, int):
            return self._get_item(index)
        elif isinstance(index, slice):
            return self._get_slice(index)
        raise TypeError(f"Record index must be int or slice, not {type(index).__name__}")

    def _get_item(self, index: int) -> T:
        if not 0 <= index < self.total_records:
            raise IndexError(f"Record index {index} out of range")
        page = index // self.per_page
        item = index % self.per_page
        return self.get_page(page)[item]  # type: ignore

    def _get_slice(self, index: slice) -> List[T]:
        # TODO: we could optimize this by not fetching unneeded pages, i.e. pages[0:1]
        # would only need the first page fetched
        records = [r for r in self]
        return records[index]

    def _get_page(self, index: int) -> Page:
        resp = self.api.get(uri=self.url, params={**self.params, **self.options.to_api(index)})
        page = ObjectMapper.tree_to_object(resp, Page)
        # NOTE: this post processing is not ideal, but works for now
        page.records = [self.record_parser(rec) for rec in page.records]  # type: ignore
        return page

    def get_page(self, index: int, force: bool = False) -> Page:
        # Validate index
        if not force:
            if not 0 <= index < self.total_pages:
                raise IndexError(f"Page index {index} out of range")
            elif index == self.page_index:
                return self.page

        # Fetch page
        self.page_index = index
        self.page = self._get_page(self.page_index)
        return self.page

    @property
    def total_pages(self) -> int:
        return ceil(self.total_records / self.per_page)

    @property
    def total_records(self) -> int:
        return self.page._metadata.totalRecords

    @property
    def per_page(self) -> int:
        return self.options.per_page


@dataclass
class TimeSeriesPageMetadata:
    count: int
    has_more: bool
    next_page_url: str
    next_record_time: int


@dataclass
class TimeSeriesPage:
    records: List[Record]
    meta: TimeSeriesPageMetadata

    def __len__(self) -> int:
        return len(self.records)

    def __iter__(self) -> Iterator[Record]:
        yield from self.records

    def __getitem__(self, index: Union[int, slice]) -> Union[Record, List[Record]]:
        return self.records[index]


class PagedTimeSeries:
    def __init__(self, api: Api, url: str, params: Optional[Dict] = None, per_page: int = 1000):
        self.api = api
        self.url = url
        self.params = params or {}
        self.params.setdefault("limit", per_page)

        # Fetch first page
        self.page_index = 0
        self.page = self._get_page(url=self.url, params=self.params)

    def __iter__(self):
        # HACK: Reset to first page
        if self.page_index != 0:
            self.page_index = 0
            self.page = self._get_page(url=self.url, params=self.params)
        yield from self.page
        while self.next_page_url:
            yield from self.get_next_page()

    def _get_page(self, url: str, params: Optional[Dict] = None) -> TimeSeriesPage:
        resp = self.api.get(url, params=params)
        page = ObjectMapper.tree_to_object(resp, TimeSeriesPage)
        # NOTE: this post processing is not ideal, but works for now
        page.records = [self._record_parser(rec) for rec in page.records]  # type: ignore
        return page

    def _record_parser(self, record: Dict) -> DataPoint:
        return Parsers.datetime(record["event_time"]), Parsers.unknown(record["value"])

    def get_next_page(self) -> TimeSeriesPage:
        if not self.next_page_url:
            raise IndexError("No next page")
        self.page_index += 1
        self.page = self._get_page(url=self.next_page_url)
        return self.page

    @property
    def next_page_url(self) -> Optional[str]:
        if not self.page.meta.next_page_url:
            return None
        return self.page.meta.next_page_url.replace(self.api.base_url, "")

    @property
    def per_page(self) -> int:
        return self.params["limit"]


@dataclass
class GQLTimeSeriesPageMetadata:
    hasNextPage: bool
    endCursor: str


@dataclass
class GQLTimeSeriesPage:
    nodes: List[Record]
    pageInfo: GQLTimeSeriesPageMetadata

    def __len__(self) -> int:
        return len(self.nodes)

    def __iter__(self) -> Iterator[Record]:
        yield from self.nodes

    def __getitem__(self, index: Union[int, slice]) -> Union[Record, List[Record]]:
        return self.nodes[index]


class PagedGQLTimeSeries:
    def __init__(
        self, api: BaseGraphService, query: str, params: Optional[Dict] = None, per_page: int = 1000
    ):
        self.api = api
        self.query = query
        self.params = params or {}
        self.params.setdefault("limit", per_page)

        # Fetch first page
        self.page_index = 0
        self.page = self._get_page()

    def __iter__(self):
        # HACK: Reset to first page
        if self.page_index != 0:
            self.page_index = 0
            self.page = self._get_page()
        yield from self.page
        while self.next_page_cursor:
            yield from self.get_next_page()

    def _get_page(self, cursor: Optional[str] = None) -> GQLTimeSeriesPage:
        params = self.params
        if cursor is not None:
            params = dict(self.params)
            params["after"] = cursor
        resp = self.api.query(self.query, variables=params)
        page = ObjectMapper.tree_to_object(resp["dataPoint"]["data"], GQLTimeSeriesPage)
        # NOTE: this post processing is not ideal, but works for now
        page.nodes = [self._record_parser(rec) for rec in page.nodes]  # type: ignore
        return page

    def _record_parser(self, record: Dict) -> DataPoint:
        return Parsers.datetime(record["time"]), Parsers.unknown(record["data"])

    def get_next_page(self) -> GQLTimeSeriesPage:
        if not self.next_page_cursor:
            raise IndexError("No next page")
        self.page_index += 1
        self.page = self._get_page(cursor=self.next_page_cursor)
        return self.page

    @property
    def next_page_cursor(self) -> Optional[str]:
        if not self.page.pageInfo.hasNextPage:
            return None
        return self.page.pageInfo.endCursor

    @property
    def per_page(self) -> int:
        return self.params["limit"]



from typing import List

import click

from contxt.cli.clients import Clients
from contxt.cli.utils import fields_option, print_table, sort_option
from contxt.models.contxt import Organization


@click.group()
def orgs() -> None:
    """Organizations."""


@orgs.command()
@fields_option(default="id, name, slug", obj=Organization)
@sort_option(default="id")
@click.pass_obj
def get(clients: Clients, fields: List[str], sort: str) -> None:
    """Get organizations"""
    items = clients.contxt.get_organizations()
    print_table(items=items, keys=fields, sort_by=sort)



import inspect
from enum import Enum
from typing import Any, Type, TypeVar, Union

T = TypeVar("T")
_PRIMITIVES = (bool, int, float, str, type(None))


class Null:
    """Null marker"""


class ObjectMapper:
    @staticmethod
    def tree_to_object(tree: object, annotation: Type[T]) -> T:
        """Loads JSON-like structure `tree` into an instance of `annotation`"""

        def failure_msg():
            return f"Failed to parse '{annotation}' from tree: {tree}"

        if tree is None or annotation == Any:
            return tree  # type: ignore
        # detecting generic types
        # note: this check is hacky, don't have a better way to detect generics
        elif type(annotation).__name__ == "_GenericAlias":
            origin = getattr(annotation, "__origin__")
            args = getattr(annotation, "__args__")
            if origin is list:
                assert isinstance(tree, list), failure_msg()
                assert len(args) == 1
                generic_param = args[0]
                return [
                    ObjectMapper.tree_to_object(item, generic_param) for item in tree  # type: ignore
                ]
            elif origin is dict:
                assert isinstance(tree, dict), failure_msg()
                assert len(args) == 2
                key_ann, value_ann = args
                return {  # type: ignore
                    ObjectMapper.tree_to_object(k, key_ann): ObjectMapper.tree_to_object(v, value_ann)
                    for k, v in tree.items()
                }
            elif origin is Union:
                assert any(t for t in args if isinstance(tree, t)), failure_msg()
                return tree  # type: ignore
        elif annotation in _PRIMITIVES:
            try:
                conv = annotation(tree)  # type: ignore
                assert conv == tree
                return conv
            except Exception as e:
                raise AssertionError(failure_msg(), e)
        elif inspect.isclass(annotation):
            if isinstance(tree, annotation):
                return tree
            elif issubclass(annotation, Enum):
                # try to map by enum entry name
                by_name = getattr(annotation, str(tree), None)
                if by_name:
                    return by_name
                # try to map by enum entry value
                assert isinstance(
                    tree, _PRIMITIVES
                ), "Only primitives are supported as Enum values. " + (failure_msg())
                return annotation(tree)  # type: ignore
            elif isinstance(tree, dict):
                # parse as regular class
                args = {}
                # get signature of constructor
                for param in inspect.signature(annotation).parameters.values():
                    raw = tree.get(param.name, Null)
                    if raw is Null:
                        assert param.default is not inspect.Parameter.empty, (
                            f"Missing value for required parameter; "
                            f"type: {annotation}, param: {param}, tree: {tree}. {failure_msg()}"
                        )
                    else:
                        args[param.name] = ObjectMapper.tree_to_object(raw, param.annotation)
                if len(tree) > len(args):
                    # there seem to be some extraneous keys
                    extraneous_keys = [k for k in tree.keys() if k not in args]
                    raise AssertionError(
                        f"Unexpected keys were found while parsing {annotation}. "
                        f"Extraneous keys: {extraneous_keys}, tree: {tree}"
                    )

                # instantiate the annotation class
                return annotation(**args)  # type: ignore

        raise NotImplementedError(failure_msg())

    load = tree_to_object

    @staticmethod
    def object_to_tree(obj):
        """Dumps `obj` to a JSON-like structure"""
        if isinstance(obj, _PRIMITIVES):
            return obj
        elif isinstance(obj, list):
            return list(map(ObjectMapper.object_to_tree, obj))
        elif isinstance(obj, Enum):
            return obj.value
        elif inspect.isclass(obj):
            return obj.__name__
        else:
            dic = obj if isinstance(obj, dict) else vars(obj)
            return {
                ObjectMapper.object_to_tree(k): ObjectMapper.object_to_tree(v) for k, v in dic.items()
            }

    dump = object_to_tree



from typing import List

import click

from contxt.cli.clients import Clients
from contxt.cli.utils import fields_option, print_item, print_table, sort_option
from contxt.models.contxt import EdgeNode, EdgeNodeGrant


@click.group("edge-nodes")
def edge_nodes() -> None:
    """Edge Nodes."""


@edge_nodes.command()
@click.argument("project_slug")
@fields_option(default="id, name, project_id, description", obj=EdgeNode)
@sort_option(default="id")
@click.pass_obj
def get(clients: Clients, project_slug: int, fields: List[str], sort: str) -> None:
    """Get edge nodes for a project"""
    items = clients.contxt_deployments.get_edge_nodes(
        organization_id=clients.org_id, project_slug=project_slug
    )
    print_table(items=items, keys=fields, sort_by=sort)


@edge_nodes.command()
@click.option("--project-slug", prompt=True)
@click.option("--name", prompt=True)
@click.option("--description", prompt=True)
@click.pass_obj
def create(
    clients: Clients,
    project_slug: str,
    name: str,
    description: str,
) -> None:
    """Create an edge node"""
    result = clients.contxt_deployments.post(
        f"{clients.org_id}/projects/{project_slug}/edgenodes",
        {
            "name": name,
            "description": description,
        },
    )
    print_item(result)


@edge_nodes.command()
@click.argument("edge_node_id")
@click.argument("target")
@click.pass_obj
def add_grant(clients: Clients, edge_node_id: str, target: int) -> None:
    """Add grant to edge node"""
    result = clients.contxt.create_edge_node_grant(
        edge_node_id=edge_node_id, to_service_instance_id=target
    )
    print_item(result)


@edge_nodes.command()
@click.argument("edge_node_id")
@fields_option(default="id, from_edge_node_id, to_service_instance_id", obj=EdgeNodeGrant)
@sort_option(default="id")
@click.pass_obj
def get_grants(clients: Clients, edge_node_id: str, fields: List[str], sort: str) -> None:
    """Get all edge node grants"""
    items = clients.contxt_deployments.get_edge_node_grants(clients.org_id, edge_node_id=edge_node_id)

    print_table(items=items, keys=fields, sort_by=sort)



from ..services.auth import AuthService
from ..utils import make_logger
from . import Auth, Token, TokenProvider

logger = make_logger(__name__)


class MachineTokenProvider(TokenProvider):
    """Concrete `TokenProvider` for a machine client, where `client_id` and
    `client_secret` serve as the identity provider.
    """

    def __init__(self, client_id: str, client_secret: str, audience: str) -> None:
        super().__init__(audience)
        self.client_id = client_id
        self.client_secret = client_secret
        self.auth_service = AuthService()

    @TokenProvider.access_token.getter  # type: ignore
    def access_token(self) -> Token:
        """Gets a valid access token for audience `audience`"""
        if self._access_token is None or self._token_expiring():
            # Token either not yet set or expiring soon, fetch one
            logger.debug(f"Fetching new access_token for {self.audience}")
            self.access_token = self.auth_service.get_oauth_token(
                self.client_id, self.client_secret, self.audience
            )["access_token"]
        return self._access_token  # type: ignore


class MachineAuth(Auth):
    """Concrete `Auth` for a machine client"""

    def get_token_provider(self, audience: str) -> MachineTokenProvider:
        """Get `MachineTokenProvider` for audience `audience`"""
        return MachineTokenProvider(self.client_id, self.client_secret, audience)



import json
from enum import Enum
from typing import Any, Dict


class DBOperation(Enum):
    INSERT = 0
    UPDATE = 1
    UPSERT = 2
    DELETE = 3

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


class CTRHeader:
    def __init__(self, forwardable=True, deduplicatable=True):
        self.forwardable = forwardable
        self.deduplicatable = deduplicatable


class ChangeTrackingRecord:
    """A JSON Serializable ndustrial.io ChangeTrackingRecord(tm) that is
    ingestable by the ndustrial ETL java/kotlin components.

    topic: Str           - the name of the canonical being ETL-ed
    version: Str         - the change table version or CDC version number (cursor position)
    operation: enum      - a DB operation (insert/update/upsert/delete) from the DBOperation enum
    source_host: Str     - the hostname/ip/table that homes the source data
    source_table: Str    - the name of the table from where the data is being extracted
    source_database: Str - the name of the db being extracted against
    tuple: dict          - key/value pairs that make up the data payload. This
                           value defaults to an empty dict and can be updated after creation
    headers: CTRHeader() - the header information that informs ETL devices what do with the CTR
    """

    def __init__(
        self,
        topic: str,
        version: int,
        operation: DBOperation,
        source_host: str,
        source_table: str,
        source_database: str,
        tuple=dict(),
        headers=CTRHeader(),
    ):
        errors = ""
        if type(topic) is not str:
            errors += "\nTopic parameter must be a string"
        if type(version) is not int:
            errors += "\nVersion parameter must be long (int)"
        if not (DBOperation.has_value(operation)):
            errors += "\nOperation parameter must be member of DBOperation enum"
        if type(source_host) is not str:
            errors += "\nsource_host parameter must be a string"
        if type(source_table) is not str:
            errors += "\nsource_host parameter must be a string"
        if type(source_database) is not str:
            errors += "\nsource_database parameter must be a string"

        if errors:
            raise AttributeError(errors)

        self.topic = topic
        self.version = version
        self.operation = operation
        self.source_host = source_host
        self.source_table = source_table
        self.source_database = source_database
        self.tuple = tuple
        self.headers = headers
        self.addlParams: Dict[str, Any] = dict()

    def getJSONString(self) -> str:
        return json.dumps(self, default=lambda x: x.__dict__)



from csv import DictWriter
from datetime import date, datetime
from enum import Enum
from json import dump, dumps
from pathlib import Path
from typing import Any, Callable, Iterable, Optional, Union

from tabulate import tabulate


class Serializer:
    """Serializer to transform a Python object to common data formats"""

    @staticmethod
    def _keys(obj):
        if isinstance(obj, dict):
            return obj.keys()
        elif isinstance(obj, (list, tuple)) and obj:
            return list(sorted(set().union(*(Serializer._keys(i) for i in obj))))
        else:
            return []

    @staticmethod
    def to_dict(obj: Any, cls_key: Union[str, None] = None, key_filter: Union[Callable, None] = None):
        """Serializes `obj` to a `dict`. To use a custom format, overload
        `obj.to_dict()`.

        :param obj: object to serialize
        :type obj: Any
        :param cls_key: if provided, set dict[key] = `cls.__name__`, defaults to None
        :param cls_key: str, optional
        :param key_filter: callable filter to keep a subset of keys, of the form
        f(key) = keep?, defaults to ignore "hidden" variables (i.e. starting with "_")
        :param key_filter: bool, optional
        :return: dictionary
        :rtype: `dict`
        """
        from contxt.models import Formatters

        # TODO: this may not return a dict but instead a list, or a native type
        # (for example, if passed an int, it will return it). this is likely
        # an unexpected behavior for callers
        # NOTE: alternative (but slower) implementations:
        # 1. d = json.dumps(d, default=lambda o: getattr(o, "__dict__", str(o)))
        # 2. d = jsonpickle.encode(d, unpicklable=False))
        # >> json.loads(d)

        def default_filter(key: str) -> bool:
            return not key.startswith("_")

        # Create default key filter (ignore _ vars)
        key_filter = key_filter or default_filter

        # Serialize
        if hasattr(obj, "to_dict"):
            # User-defined method
            return obj.to_dict()
        if isinstance(obj, dict):
            # Dictionary
            return {k: Serializer.to_dict(v, cls_key=cls_key) for k, v in obj.items()}
        elif isinstance(obj, datetime):
            # Datetime
            return Formatters.datetime(obj)
        elif isinstance(obj, date):
            # Date
            return Formatters.date(obj)
        elif isinstance(obj, Enum):
            return obj.value
        elif hasattr(obj, "_ast"):
            # Abstract syntax tree
            return Serializer.to_dict(obj._ast())
        elif hasattr(obj, "__dict__"):
            # General object
            d = {
                k: Serializer.to_dict(v, cls_key)
                for k, v in obj.__dict__.items()
                if not callable(v) and key_filter(k)
            }
            if cls_key is not None and hasattr(obj, "__class__"):
                d[cls_key] = obj.__class__.__name__
            return d
        elif hasattr(obj, "__iter__") and not isinstance(obj, str):
            # Iterables (except strings)
            return [Serializer.to_dict(v, cls_key) for v in obj]
        else:
            # Fallback to self
            return obj

    @staticmethod
    def to_json(obj: Any, path: Optional[Path] = None, **kwargs):
        """Serializes `obj` to JSON.

        :param obj: object to serialize
        :type obj: Any
        :param path: path for dumping, defaults to None
        :param path: Path, optional
        :return: JSON
        :rtype: str
        """
        # Set default args
        kwargs.setdefault("indent", "\t")
        kwargs.setdefault("sort_keys", True)

        # Dump
        d = Serializer.to_dict(obj)
        if path:
            with path.open("w") as f:
                return dump(d, f, **kwargs)
        return dumps(d, **kwargs)

    @staticmethod
    def to_pretty_cli(obj: Any, **kwargs):
        d = Serializer.to_dict(obj)

        nested_sections = {}
        for key, value in d.items():
            if isinstance(value, (list, tuple, dict)):
                nested_sections[key] = value

        # delete the values out of the main section that we're plucking out to print below
        printed = "\n====" + type(obj).__name__ + "====\n"
        for key in nested_sections:
            del d[key]
        printed += Serializer.to_table(d) + "\n"

        for key, value in nested_sections.items():
            printed += "\n" + key.capitalize() + "\n"
            printed += Serializer.to_table(value)
            printed += "\n"

        return printed

    @staticmethod
    def to_table(obj: Any, path: Optional[Path] = None, exclude_keys=None, sort_by=None, **kwargs):
        """Serializes `obj` to a table.

        :param obj: object to serialize
        :type obj: Any
        :param path: path for dumping, defaults to None
        :param path: Path, optional
        :return: table
        :rtype: tabulate
        """

        if exclude_keys is None:
            exclude_keys = []

        d = Serializer.to_dict(obj)

        # TODO: The above may be a list, so handle it here
        if not isinstance(d, (list, tuple)):
            d = [d]

        filtered = [{k: v for k, v in item.items() if k not in exclude_keys} for item in d]

        if sort_by is not None:
            filtered = sorted(filtered, key=lambda item: item[sort_by])

        table = tabulate(filtered, headers="keys", **kwargs)
        if path:
            with path.open("w") as f:
                return f.write(table)
        return table

    @staticmethod
    def to_csv(obj: Any, path: Path, header: bool = True, **kwargs):
        """Serializes `obj` to CSV.

        :param obj: object to serialize
        :type obj: Any
        :param path: path to export
        :type path: Path
        :param header: write header row, defaults to True
        :param header: bool, optional
        """

        # Get dict
        d = Serializer.to_dict(obj)

        # Write
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("w") as f:
            writer = DictWriter(f, fieldnames=Serializer._keys(d), **kwargs)
            if header:
                writer.writeheader()
            if isinstance(d, (list, tuple)):
                # Write list
                writer.writerows(d)
            else:
                # TODO: may not actually be a dict
                # Write dict
                writer.writerow(d)

    @staticmethod
    def to_file(
        obj: Any, path: Optional[Path] = None, valid_exts: Iterable[str] = (".csv", ".json", ".txt")
    ):
        """Write an object to a file (or stdout).

        If path is not provided, `obj` is sent to stdout. Otherwise, the file
        extension is used to determine the data format of `obj` before exporting.
        Supported file extensions are csv (calls `Serializer.to_csv`),
        json (calls `Serializer.to_json`), or txt (calls `Serializer.to_table`).

        :param obj: object to output
        :type obj: Any
        :param path: path to export, defaults to None
        :param path: Optional[Path], optional
        :param valid_exts: supported file extensions, defaults to
        (".csv", ".json", ".txt")
        :param valid_exts: Optional[List[str]], optional
        """

        # If path unspecified, default to stdout
        if path is None:
            obj_to_print = (
                Serializer.to_table(obj) if isinstance(obj, (list, tuple)) else Serializer.to_json(obj)
            )
            print(obj_to_print)
            return

        # Validate file extension
        if path.suffix not in valid_exts:
            raise RuntimeError(
                f"Unsupported filetype: '{path.suffix}'. Choose from {', '.join(valid_exts)}"
            )

        # Dump to file
        if path.suffix == ".csv":
            Serializer.to_csv(obj, path)
        elif path.suffix == ".json":
            Serializer.to_json(obj, path)
        else:
            Serializer.to_table(obj, path)



from abc import ABC, abstractmethod
from os import environ
from typing import Any, Dict, Optional, Tuple

from .auth.machine import MachineAuth
from .models.contxt import Config
from .services import ContxtService


class BaseWorker(ABC):
    """Abstract base class for a worker. Overload this class to implement `do_work()`"""

    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        environment_id: Optional[str] = None,
    ) -> None:
        self.client_id = client_id or environ["CLIENT_ID"]
        self.client_secret = client_secret or environ["CLIENT_SECRET"]
        self.environment_id = environment_id or environ.get("ENVIRONMENT_ID")
        self.auth = MachineAuth(self.client_id, self.client_secret)
        self.contxt_service = ContxtService(self.auth)

        # Retrieve configuration from Contxt (if one exists)
        self.config, self.config_values = self._init_configuration()

    def _init_configuration(self) -> Tuple[Optional[Config], Dict[str, Any]]:
        if not self.environment_id:
            # No env specified, no values to fetch
            return None, {}

        # Fetch contxt config
        config = self.contxt_service.get_config_for_client(self.client_id, self.environment_id)

        # Cache values in a simple dict for easy consumption
        config_values = {v.key: v.value for v in config.config_values} if config else {}

        return config, config_values

    @abstractmethod
    def do_work(self) -> None:
        """Main method"""



from dataclasses import dataclass
from datetime import datetime
from json import loads
from typing import ClassVar, Optional

from . import ApiField, ApiObject, Parsers


@dataclass
class EventType(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("name", creatable=True),
        ApiField("slug", creatable=True),
        ApiField("description", creatable=True),
        ApiField("client_id", creatable=True),
        ApiField("level", data_type=int, creatable=True),
        ApiField("is_ongoing_event", data_type=bool, optional=True),
        ApiField("is_realtime_enabled", data_type=bool, creatable=True),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    name: str
    slug: str
    description: str
    client_id: str
    level: int
    id: Optional[str] = None
    is_ongoing_event: Optional[bool] = None
    is_realtime_enabled: Optional[bool] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


@dataclass
class EventDefinition(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("event_id"),
        ApiField("description"),
        ApiField("parameters", data_type=lambda o: loads(o.replace('"', '"'))),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    event_id: str
    description: str
    parameters: dict
    created_at: datetime
    updated_at: datetime
    id: Optional[str] = None
    human_readable_parameters: Optional[str] = None


@dataclass
class Owner(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("first_name"),
        ApiField("last_name"),
        ApiField("email"),
        ApiField("is_machine_user", data_type=bool, optional=True),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
    )

    id: str
    first_name: str
    last_name: str
    email: str
    is_machine_user: bool
    created_at: datetime
    updated_at: datetime


@dataclass
class Event(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("name", creatable=True),
        ApiField("event_type_id", creatable=True),
        ApiField("EventType", attr_key="event_type", data_type=EventType, optional=True),
        ApiField("organization_id", creatable=True),
        ApiField("facility_id", data_type=int, creatable=True),
        ApiField("owner_id"),
        ApiField("Owner", attr_key="owner", data_type=Owner, optional=True),
        ApiField("topic_arn", optional=True),
        ApiField("allow_others_to_trigger", data_type=bool, creatable=True),
        ApiField("is_public", data_type=bool, creatable=True),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
        ApiField("deleted_at", data_type=Parsers.datetime),
    )

    name: str
    event_type_id: str
    organization_id: str
    allow_others_to_trigger: bool
    is_public: bool
    id: Optional[str] = None
    facility_id: Optional[int] = None
    owner_id: Optional[str] = None
    owner: Optional[Owner] = None
    event_type: Optional[EventType] = None
    topic_arn: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    deleted_at: Optional[datetime] = None


@dataclass
class TriggeredEvent(ApiObject):
    _api_fields: ClassVar = (
        ApiField("id"),
        ApiField("event_id", creatable=True),
        ApiField("Event", data_type=Event, attr_key="event", optional=True),
        ApiField("owner_id"),
        ApiField("Owner", data_type=dict, attr_key="owner", optional=True),
        ApiField("is_public", data_type=bool),
        ApiField("data", creatable=True, updatable=True),
        ApiField("trigger_start_at", data_type=Parsers.datetime, creatable=True, updatable=True),
        ApiField("trigger_end_at", data_type=Parsers.datetime, creatable=True, updatable=True),
        ApiField("created_at", data_type=Parsers.datetime),
        ApiField("updated_at", data_type=Parsers.datetime),
        ApiField("deleted_at", data_type=Parsers.datetime),
    )

    event_id: str
    trigger_start_at: datetime
    id: Optional[str] = None
    owner_id: Optional[str] = None
    owner: Optional[Owner] = None
    event: Optional[Event] = None
    is_public: Optional[bool] = None
    trigger_end_at: Optional[datetime] = None
    data: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    deleted_at: Optional[datetime] = None



from typing import List, Optional

import click

from contxt.cli.clients import Clients
from contxt.cli.utils import fields_option, print_item, print_table, sort_option
from contxt.models.contxt import ServiceInstance, ServiceInstanceGrant


@click.group()
def service_instances() -> None:
    """Service Instances."""


@service_instances.command()
@click.option("--service-id")
@click.pass_obj
@fields_option(default="id, slug, service_id, description", obj=ServiceInstance)
@sort_option(default="slug")
def get(clients: Clients, service_id: Optional[str], fields: List[str], sort: str) -> None:
    """Get service instance(s)"""
    items = (
        clients.contxt_deployments.get(f"{clients.org_id}/services/{service_id}/service_instances")
        if service_id
        else clients.contxt_deployments.get(f"{clients.org_id}/service_instances")
    )
    print_table(items, keys=fields, sort_by=sort)


@service_instances.command()
@click.option("--service-id", prompt=True)
@click.option("--project-environment-id", required=False)
@click.option("--name", prompt=True)
@click.option("--slug", prompt=True)
@click.option("--descriptor", prompt=True)
@click.option("--service-type", type=click.Choice(["API", "Application", "Worker"]), prompt=True)
@click.pass_obj
def create(
    clients: Clients,
    service_id: str,
    project_environment_id: Optional[str],
    name: str,
    slug: str,
    descriptor: str,
    service_type: str,
) -> None:
    """Create service instances"""
    json = {
        "name": name,
        "project_environment_id": project_environment_id,
        "slug": slug,
        "descriptor": descriptor,
        "service_type": service_type,
    }
    json = {k: v for k, v in json.items() if v is not None}
    result = clients.contxt_deployments.post(
        f"{clients.org_id}/services/{service_id}/service_instances", json
    )
    print_item(result)


@service_instances.command()
@click.argument("id")
@click.pass_obj
def delete(clients: Clients, id: str) -> None:
    """Delete service instance"""
    clients.contxt_deployments.delete(f"{clients.org_id}/service_instances/{id}")


@service_instances.command()
@click.argument("from_service_instance_id")
@click.argument("to_service_instance_id")
@click.pass_obj
def add_grant(clients: Clients, from_service_instance_id: int, to_service_instance_id: int) -> None:
    """Add grant to service instance"""
    json = {"to_service_instance_id": to_service_instance_id}
    result = clients.contxt_deployments.post(
        f"{clients.org_id}/service_instances/{from_service_instance_id}/grants", json
    )
    print_item(result)


@service_instances.command()
@click.argument("service_instance_id")
@fields_option(default="id, from_service_instance_id, to_service_instance_id", obj=ServiceInstanceGrant)
@sort_option(default="id")
@click.pass_obj
def get_grants(clients: Clients, service_instance_id: str, fields: List[str], sort: str) -> None:
    """Get all service instance grants"""
    items = clients.contxt_deployments.get_service_instance_grants(
        clients.org_id, service_instance_id=service_instance_id
    )
    print_table(items=items, keys=fields, sort_by=sort)


@service_instances.command()
@click.option("--service_instance_id", prompt=True)
@click.option("--label", prompt=True)
@click.option("--description", prompt=True)
@click.pass_obj
def add_scope(clients: Clients, service_instance_id: int, label: str, description: str) -> None:
    """Create service instance scope"""
    json = {"label": label, "description": description}
    json = {k: v for k, v in json.items() if v is not None}
    result = clients.contxt_deployments.post(
        f"{clients.org_id}/service_instances/{service_instance_id}/scopes", json
    )
    print_item(result)


@service_instances.command()
@click.option("--service_instance_id", prompt=True)
@click.option(
    "--url_type",
    type=click.Choice(
        ["allowed_clients", "allowed_logout_urls", "allowed_origins", "callbacks", "web_origins"]
    ),
    prompt=True,
)
@click.option("--url", prompt=True)
@click.pass_obj
def add_url(clients: Clients, service_instance_id: int, url_type: str, url: str) -> None:
    """Add service instance url"""
    json = {"url_type": url_type, "url": url}
    json = {k: v for k, v in json.items() if v is not None}
    result = clients.contxt_deployments.post(
        f"{clients.org_id}/service_instances/{service_instance_id}/urls", json
    )
    print_item(result)



"""Auth for API clients"""

from abc import ABC, abstractmethod
from datetime import datetime, timezone
from typing import Any, Dict, Optional

from jwt import decode

Token = str
DecodedToken = Dict[str, Any]


# TODO: Replace with https://github.com/requests/requests-oauthlib
class TokenProvider(ABC):
    """An abstact base class to provide an unexpired `access_token` from the source
    client to the target client, defined by `audience`.

    Overload this class to implement `access_token`.
    """

    def __init__(self, audience: str) -> None:
        self.audience = audience
        self._access_token: Optional[Token] = None
        self._access_token_decoded: Optional[DecodedToken] = None

    def _token_expiring(self, within: int = 60) -> bool:
        """Returns if `access_token` is within `within` seconds of expiring.
        Returns `False` if `exp` is not in the token's claims."""
        expiration = self.decoded_access_token.get("exp")
        if not expiration:
            return False
        return expiration <= datetime.now(timezone.utc).timestamp() + within

    @property
    @abstractmethod
    def access_token(self) -> str:
        """Gets a valid access token for audience `audience`"""

    @access_token.setter
    def access_token(self, value: Token) -> None:
        """Sets both the access token and the decoded access token"""
        self._access_token = value
        self._access_token_decoded = decode(self._access_token, options={"verify_signature": False})

    @property
    def decoded_access_token(self) -> DecodedToken:
        """Gets the decoded access token"""
        if self._access_token_decoded is None:
            # Token not yet set, fetch it now
            self.access_token
        return self._access_token_decoded  # type: ignore

    def reset(self) -> None:
        self._access_token = None
        self._access_token_decoded = None


class Auth(ABC):
    """An abstract base class for a client defined by `client_id` and `client_secret`.
    To authenticate to another client, use `get_token_provider()`.

    Overload this class to implement `get_token_provider()`.
    """

    def __init__(self, client_id: str, client_secret: str) -> None:
        self.client_id = client_id
        self.client_secret = client_secret

    @abstractmethod
    def get_token_provider(self, audience: str) -> TokenProvider:
        """Get `TokenProvider` for audience `audience`"""



from typing import Iterable, List, Optional

from ..auth import Auth
from ..models.events import Event, EventDefinition, EventType, TriggeredEvent
from .api import ApiEnvironment, ConfiguredApi
from .pagination import PagedRecords, PageOptions


class EventsService(ConfiguredApi):
    """Events API client"""

    _envs = (
        ApiEnvironment(
            name="production",
            base_url="https://{tenant}.api.ndustrial.io/rest/events/v1",
            client_id="7jzwfE20O2XZ4aq3cO1wmk63G9GzNc8j",
        ),
        ApiEnvironment(
            name="staging",
            base_url="https://{tenant}.api.staging.ndustrial.io/rest/events/v1",
            client_id="dn4MaocJFdKtsBy9sFFaTeuJWL1nt5xu",
        ),
    )

    def __init__(self, auth: Auth, org_slug: str, env: str = "production", **kwargs) -> None:
        super().__init__(env=env, auth=auth, **kwargs)
        self.base_url = self.base_url.format(tenant=org_slug)

    def set_human_readable_parameters(self, event_definition: EventDefinition) -> None:
        statement = ""
        for k, v in event_definition.parameters.items():
            if k == "$chain":
                for d1, d2 in zip(v[:-1], v[1:]):
                    event1 = self.get_event(d1.get("event_id"))
                    event2 = self.get_event(d2.get("event_id"))
                    mins = d1.get("overlap_variance") / 60
                    statement += f"Event {event1.name} overlaps with {event2.name} within {mins} min "
        event_definition.human_readable_parameters = statement

    def get_event_types(self, page_options: Optional[PageOptions] = None) -> Iterable[EventType]:
        return PagedRecords(
            api=self, url="types", options=page_options, record_parser=EventType.from_api
        )

    def create_event_type(self, event_type: EventType) -> EventType:
        resp = self.post("types", data=event_type.post())
        return EventType.from_api(resp)

    def get_event(self, event_id: str) -> Event:
        resp = self.get(f"events/{event_id}")
        return Event.from_api(resp)

    def create_event(self, event: Event) -> Event:
        resp = self.post("events", data=event.post())
        return Event.from_api(resp)

    def update_event(self, event: Event) -> None:
        self.put(f"events/{event.id}", data=event.put())

    def delete_event(self, event: Event) -> None:
        self.delete(f"events/{event.id}")

    def get_events_for_type(
        self, event_type_id: str, page_options: Optional[PageOptions] = None
    ) -> Iterable[Event]:
        return PagedRecords(
            api=self,
            url=f"types/{event_type_id}/events",
            options=page_options,
            record_parser=Event.from_api,
        )

    def get_events_for_client(self, client_id: str) -> List[Event]:
        resp = self.get(f"clients/{client_id}/events")
        return [Event.from_api(rec) for rec in resp]

    def get_event_definition(self, event_id: str) -> EventDefinition:
        resp = self.get(f"events/{event_id}/definition")
        return EventDefinition.from_api(resp)

    def create_triggered_event(self, triggered_event: TriggeredEvent) -> TriggeredEvent:
        resp = self.post(f"events/{triggered_event.event_id}/trigger", data=triggered_event.post())
        return TriggeredEvent.from_api(resp)

    def get_triggered_event(self, triggered_event_id: str) -> TriggeredEvent:
        resp = self.get(f"triggered/{triggered_event_id}")
        return TriggeredEvent.from_api(resp)

    def update_triggered_event(self, triggered_event: TriggeredEvent) -> None:
        self.put(f"triggered/{triggered_event.id}", data=triggered_event.put())

    def delete_triggered_event(self, triggered_event: TriggeredEvent) -> None:
        self.delete(f"triggered/{triggered_event.id}")

    def get_triggered_events_for_event(
        self, event_id: str, page_options: Optional[PageOptions] = None
    ) -> Iterable[TriggeredEvent]:
        return PagedRecords(
            api=self,
            url=f"events/{event_id}/triggered",
            options=page_options,
            record_parser=TriggeredEvent.from_api,
        )

    def get_triggered_events_for_field(self, field_id: str) -> List[TriggeredEvent]:
        resp = self.get(f"fields/{field_id}/triggered")
        return [TriggeredEvent.from_api(rec) for rec in resp]
